/**
*Buzz Chess Engine
*eval.c
*
*I would like to note that in this eval I used, although not exactly,
*many of the principles suggested by the correspondence chess champion
*Hans Berliner.
*
*Copyright (C) 2008 Pradu Kannan. All rights reserved.
**/
#ifdef ANDRES_EVAL
#if defined(_WIN32) || defined(_WIN64)
	#define VC_EXTRALEAN
	#define WIN32_LEAN_AND_MEAN
	//#define NOMINMAX
	#include <windows.h>

#endif
#endif

#include "eval.h"
#include "board.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "thread.h"
#include "movegen.h"
#include "bitinstructions.h"
#include "egbb.h"
#ifdef max
	#undef max
#endif
static INLINE int max(int a, int b)
{
	if(a>b) return a;
	return b;
}

#ifdef min
	#undef min
#endif
static INLINE int min(int a, int b)
{
	if(a<b) return a;
	return b;
}
#ifdef DEBUG_EVAL
#define DEBUG_EVAL_PRINT(TEXT) \
{ \
	int debug_op[2] = {op[WHITE]-debugop[WHITE], op[BLACK]-debugop[BLACK]}; \
	int debug_eg[2] = {eg[WHITE]-debugeg[WHITE], eg[BLACK]-debugeg[BLACK]}; \
	int debug_compiled[2] = { \
	SCALE_SCORE(debug_op[WHITE],debug_eg[WHITE],pieceMaterial[BLACK]), \
	SCALE_SCORE(debug_op[BLACK],debug_eg[BLACK],pieceMaterial[WHITE]) \
	}; \
	printf(TEXT " = %d\n",score-debugscore+debug_compiled[pos->side]-debug_compiled[pos->xside]); \
	printf("       Opening = [%4d,%4d] = %4d\n",debug_op[WHITE],debug_op[BLACK],debug_op[pos->side]-debug_op[pos->xside]); \
	printf("       Endgame = [%4d,%4d] = %4d\n",debug_eg[WHITE],debug_eg[BLACK],debug_eg[pos->side]-debug_eg[pos->xside]); \
	debugop[WHITE]=op[WHITE]; \
	debugop[BLACK]=op[BLACK]; \
	debugeg[WHITE]=eg[WHITE]; \
	debugeg[BLACK]=eg[BLACK]; \
	debugscore=score; \
}
#else
#define DEBUG_EVAL_PRINT(TEXT) /*Nothing*/
#endif

//Used for SEE and other search related stuff
int pieceValue[7]={Pval,Nval,Bval,Rval,Qval,Kval,0};

/*Ideas

Safe-Mobility
PST-Pawn Eval

Knights are easily chased away with pawn moves. Therefore it is important
to spot "holes" in the enemy position where a knight cannot be attacked,
because the pawns have already moved past. Once such a hole is identified,
a knight should be maneuvered to that location. An unchallengeable knight
on the fifth row is a strong asset, and a supported knight on the sixth
row usually decides the game.

An endgame in which the parties have bishops that live on different colours
is almost always drawn, even if one side is two pawns ahead.
simply reduce the material score
more when less material on board

Queens are the most powerful pieces in a chess game. Queens are extremely
versatile, and can threaten many pieces at once. For this reason, checkmates
involving the queen are much easier to achieve than those without her. Because
the loss of a queen usually results in the loss of the game, it is generally
wise to wait to develop a queen until after the knights and bishops have been
developed.

In the endgame, the king becomes a strong piece. With reduced material, mate
is not an immediate concern anymore, and the king should be moved towards the
center of the board.

Initiative Eval

Should mobility weight be the same for every piece?

An Isolated pawn is a pawn that has no friendly pawns on adjacent files.
Such a pawn cannot be defended by another pawn and therefore needs to be
defended by pieces instead. Nonetheless an isolated pawn in the middlegame
isn't always considered a handicap, as the player with the isolated pawn has
open files for the rooks on either side of the pawn. Also, in typical
isolated pawn positions, the isolated pawn can be used to break open the
centre with a well timed advance. However, once the endgame is reached, an
isolated pawn is more likely to become a serious weakness.

If you have a passed pawn, you may want to trade off pieces to protect it.
*/
#ifndef ANDRES_EVAL
//Material Piece Square Tables

//Ratios from Hans Berliner's system

int P_op_PST[64] = {
	 P_op*90/100,  P_op*95/100, P_op*105/100, P_op*110/100, P_op*110/100, P_op*105/100,   P_op*95/100,  P_op*90/100,
	 P_op*90/100,  P_op*95/100, P_op*105/100, P_op*110/100, P_op*110/100, P_op*105/100,   P_op*95/100,  P_op*90/100,
	 P_op*90/100,  P_op*95/100, P_op*105/100, P_op*115/100, P_op*115/100, P_op*105/100,   P_op*95/100,  P_op*90/100,
	 P_op*90/100,  P_op*95/100, P_op*110/100, P_op*120/100, P_op*120/100, P_op*110/100,   P_op*95/100,  P_op*90/100,
	 P_op*90/100,  P_op*95/100, P_op*105/100, P_op*115/100, P_op*115/100, P_op*105/100,   P_op*95/100,  P_op*90/100,
	 P_op*90/100,  P_op*95/100, P_op*105/100, P_op*110/100, P_op*110/100, P_op*105/100,   P_op*95/100,  P_op*90/100,
	 P_op*90/100,  P_op*95/100, P_op*105/100, P_op*110/100, P_op*110/100, P_op*105/100,   P_op*95/100,  P_op*90/100,
	 P_op*90/100,  P_op*95/100, P_op*105/100, P_op*110/100, P_op*110/100, P_op*105/100,   P_op*95/100,  P_op*90/100,
};

int P_eg_PST[64] = {
	P_eg*120/100, P_eg*105/100,  P_eg*95/100,  P_eg*90/100,  P_eg*90/100,  P_eg*95/100,  P_eg*105/100, P_eg*120/100,
	P_eg*120/100, P_eg*105/100,  P_eg*95/100,  P_eg*90/100,  P_eg*90/100,  P_eg*95/100,  P_eg*105/100, P_eg*120/100,
	P_eg*120/100, P_eg*105/100,  P_eg*95/100,  P_eg*90/100,  P_eg*90/100,  P_eg*95/100,  P_eg*105/100, P_eg*120/100,
	P_eg*125/100, P_eg*110/100, P_eg*100/100,  P_eg*95/100,  P_eg*95/100, P_eg*100/100,  P_eg*110/100, P_eg*125/100,
	P_eg*133/100, P_eg*117/100, P_eg*107/100, P_eg*100/100, P_eg*100/100, P_eg*107/100,  P_eg*117/100, P_eg*133/100,
	P_eg*145/100, P_eg*129/100, P_eg*116/100, P_eg*105/100, P_eg*105/100, P_eg*116/100,  P_eg*129/100, P_eg*145/100,
	P_eg*145/100, P_eg*129/100, P_eg*116/100, P_eg*105/100, P_eg*105/100, P_eg*116/100,  P_eg*129/100, P_eg*145/100,
	P_eg*145/100, P_eg*129/100, P_eg*116/100, P_eg*105/100, P_eg*105/100, P_eg*116/100,  P_eg*129/100, P_eg*145/100
};

//piece square table for pieces - similar to Toga's
int K_op_PST[64] = {
	 4*K_op_PSTW,  4*K_op_PSTW,  3*K_op_PSTW,  1*K_op_PSTW,  1*K_op_PSTW,  3*K_op_PSTW,  4*K_op_PSTW,  4*K_op_PSTW,
	 3*K_op_PSTW,  4*K_op_PSTW,  2*K_op_PSTW,            0,            0,  2*K_op_PSTW,  4*K_op_PSTW,  3*K_op_PSTW,
	 1*K_op_PSTW,  2*K_op_PSTW,            0, -2*K_op_PSTW, -2*K_op_PSTW,            0,  2*K_op_PSTW,  1*K_op_PSTW,
               0,  1*K_op_PSTW, -1*K_op_PSTW, -3*K_op_PSTW, -3*K_op_PSTW, -1*K_op_PSTW,  1*K_op_PSTW,            0,
	-1*K_op_PSTW,            0, -2*K_op_PSTW, -4*K_op_PSTW, -4*K_op_PSTW, -2*K_op_PSTW,            0, -1*K_op_PSTW,
	-2*K_op_PSTW, -1*K_op_PSTW, -3*K_op_PSTW, -5*K_op_PSTW, -5*K_op_PSTW, -3*K_op_PSTW, -1*K_op_PSTW, -2*K_op_PSTW,
	-3*K_op_PSTW, -2*K_op_PSTW, -4*K_op_PSTW, -6*K_op_PSTW, -6*K_op_PSTW, -4*K_op_PSTW, -2*K_op_PSTW, -3*K_op_PSTW,
	-4*K_op_PSTW, -3*K_op_PSTW, -5*K_op_PSTW, -7*K_op_PSTW, -7*K_op_PSTW, -5*K_op_PSTW, -3*K_op_PSTW, -4*K_op_PSTW
};

int K_eg_PST[64] = {
	-K_eg_PSTW*6, -K_eg_PSTW*4, -K_eg_PSTW*3, -K_eg_PSTW*2, -K_eg_PSTW*2, -K_eg_PSTW*3, -K_eg_PSTW*4, -K_eg_PSTW*6,
	-K_eg_PSTW*4, -K_eg_PSTW*2, -K_eg_PSTW*1,            0,            0, -K_eg_PSTW*1, -K_eg_PSTW*2, -K_eg_PSTW*4,
	-K_eg_PSTW*3, -K_eg_PSTW*1,            0,  K_eg_PSTW*1,  K_eg_PSTW*1,            0, -K_eg_PSTW*1, -K_eg_PSTW*3,
	-K_eg_PSTW*2,            0,  K_eg_PSTW*1,  K_eg_PSTW*2,  K_eg_PSTW*2,  K_eg_PSTW*1,            0, -K_eg_PSTW*2,
	-K_eg_PSTW*2,            0,  K_eg_PSTW*1,  K_eg_PSTW*2,  K_eg_PSTW*2,  K_eg_PSTW*1,            0, -K_eg_PSTW*2,
	-K_eg_PSTW*3, -K_eg_PSTW*1,            0,  K_eg_PSTW*1,  K_eg_PSTW*1,            0, -K_eg_PSTW*1, -K_eg_PSTW*3,
	-K_eg_PSTW*4, -K_eg_PSTW*2, -K_eg_PSTW*1,            0,            0, -K_eg_PSTW*1, -K_eg_PSTW*2, -K_eg_PSTW*4,
	-K_eg_PSTW*6, -K_eg_PSTW*4, -K_eg_PSTW*3, -K_eg_PSTW*2, -K_eg_PSTW*2, -K_eg_PSTW*3, -K_eg_PSTW*4, -K_eg_PSTW*6
};

//Rank Tables for passed and halfpassed pawns - similar to Toga
int RankBonus[8] = {0, 0, 0, 26, 77, 154, 256, 256}; //Divide by 256

//Another RankBonus
int RankBonus2[8] = {184, 184, 184, 202, 220, 238, 256, 256}; //Divide by 256

//pawn structure piece square tables for opening
//endgame pawn structure scores are the opening scores multiplied by P_eg/P_op.
int passedPawnPST[64] = {
	 0,  0,  0,  0,  0,  0,  0,  0,
	27, 26, 24, 23, 23, 24, 26, 27,
	27, 26, 24, 23, 23, 24, 26, 27,
	36, 35, 33, 31, 31, 33, 35, 36,
	49, 47, 44, 42, 42, 44, 47, 49,
	66, 63, 60, 56, 56, 60, 63, 66,
	90, 85, 81, 76, 76, 81, 85, 90,
	 0,  0,  0,  0,  0,  0,  0,  0
};

int halfpassedPawnPST[64] = {
	 0,  0,  0,  0,  0,  0,  0,  0,
	 8,  8,  7,  7,  7,  7,  8,  8,
	 8,  8,  7,  7,  7,  7,  8,  8,
	11, 10, 10,  9,  9, 10, 10, 11,
	15, 14, 13, 12, 12, 13, 14, 15,
	20, 19, 17, 16, 16, 17, 19, 20,
	27, 25, 24, 22, 22, 24, 25, 27,
	 0,  0,  0,  0,  0,  0,  0,  0
};

//Kingsafety Weight Table
//int KSWT[]=KS_NUMATTACKERS_WEIGHT_TABLE;

//Lone King Mate Recognizer PSTs
int recog_R_Q_BB[64] = {
	9, 8, 7, 6, 6, 7, 8, 9,
	8, 5, 4, 3, 3, 4, 5, 8,
	7, 4, 2, 1, 1, 2, 4, 7,
	6, 3, 1, 0, 0, 1, 3, 6,
	6, 3, 1, 0, 0, 1, 3, 6,
	7, 4, 2, 1, 1, 2, 4, 7,
	8, 5, 4, 3, 3, 4, 5, 8,
	9, 8, 7, 6, 6, 7, 8, 9,
};

//recognizer for BN black bishop mate
//use the flip pst for the white bishop mate
int recog_BN[64] = {
	10,  9,  8,  7,  6,  5,  4,  3,
	 9,  7,  6,  5,  4,  3,  2,  4,
	 8,  6,  4,  3,  2,  1,  3,  5,
	 7,  5,  3,  1,  0,  2,  4,  6,
	 6,  4,  2,  0,  1,  3,  5,  7,
	 5,  3,  1,  2,  3,  4,  6,  8,
	 4,  2,  3,  4,  5,  6,  7,  9,
	 3,  4,  5,  6,  7,  8,  9, 10,
};

//evauation that takes into account recognition of the type of position
int eval(const board* pos, int alpha, int beta)
{
	/***********************************************************************/
	/*                         Lone King Recognizer                        */
	/***********************************************************************/
	/*if(!extractMatSig(pos->recogsig,WHITE)) //lone white king
	{
		bool recognized=false;
		//scoring for black
		int score = RECOG_WIN+materialBLACK(pos);
		unsigned int msig = extractMatSig(pos->recogsig,BLACK);
		//Queen/Rook/Two Bishops
		if(msigHasPiece(msig,Q) || msigHasPiece(msig,R) || (pos->Pieces[B]&WHITESQUARES && pos->Pieces[B]&BLACKSQUARES))
		{
			recognized=true;
			score+=recog_R_Q_BB[pos->KingPos[WHITE]]-distance(pos->KingPos[WHITE],pos->KingPos[BLACK]);
		}
		//Bishop+Knight
		else if(msigHasPiece(msig,B) && msigHasPiece(msig,N))
		{
			recognized=true;
			if(pos->Pieces[B]&BLACKSQUARES)
				score+=recog_BN[pos->KingPos[WHITE]];
			else
				score+=recog_BN[FLIPV(pos->KingPos[WHITE])];
			score-=distance(pos->KingPos[WHITE],pos->KingPos[BLACK]);
		}
		//multiply score by -1 if the current side is white
		if(recognized)
		{
			if(pos->side==WHITE) return -score;
			else return score;
		}
	}
	else if(!extractMatSig(pos->recogsig,BLACK)) //lone black king
	{
		bool recognized=false;
		int score = RECOG_WIN+materialWHITE(pos);
		unsigned int msig = extractMatSig(pos->recogsig,WHITE);
		//Queen/Rook/Two Bishops
		if(msigHasPiece(msig,Q) || msigHasPiece(msig,R) || (pos->Pieces[B]&WHITESQUARES && pos->Pieces[B]&BLACKSQUARES))
		{
			recognized=true;
			score+=recog_R_Q_BB[pos->KingPos[BLACK]]-distance(pos->KingPos[BLACK],pos->KingPos[WHITE]);
		}
		//Bishop+Knight
		else if(msigHasPiece(msig,B) && msigHasPiece(msig,N))
		{
			recognized=true;
			if(pos->Pieces[B]&BLACKSQUARES)
				score+=recog_BN[pos->KingPos[BLACK]];
			else
				score+=recog_BN[FLIPV(pos->KingPos[BLACK])];
			score-=distance(pos->KingPos[BLACK],pos->KingPos[WHITE]);
		}
		//multiply score by -1 if the current side is black
		if(recognized)
		{
			if(pos->side==BLACK) return -score;
			else return score;
		}
	}*/
#ifdef ZERO_EVAL
	return 0;
#elif defined(RAND_EVAL)
	return (pos->hashkey&(255|toBit(63)));
#else
	return fulleval(pos,alpha,beta);
#endif
}

//pure eval
int fulleval(const board* pos, int alpha, int beta)
{
	/***********************************************************************/
	/*                         Variable Declaration                        */
	/***********************************************************************/

	//Evaluation Scores
	//phase-independent score
	int score = 0;

	//phase-dependent score
	//op[side], eg[side]
	//white's score gets mixed according to black's piece material
	//black's score gets mixed according to white's piece material
	int op[2] = {0,0};
	int eg[2] = {0,0};


	#ifdef DEBUG_EVAL
		int debugscore = 0;
		int debugop[2] = {0,0};
		int debugeg[2] = {0,0};
	#endif

	//the piece bitboards for each side
	U64 Pawns[2]={piecesWHITE(*pos,P),piecesBLACK(*pos,P)};
	U64 Knights[2]={piecesWHITE(*pos,N),piecesBLACK(*pos,N)};
	U64 Bishops[2]={piecesWHITE(*pos,B),piecesBLACK(*pos,B)};
	U64 Rooks[2]={piecesWHITE(*pos,R),piecesBLACK(*pos,R)};
	U64 Queens[2]={piecesWHITE(*pos,Q),piecesBLACK(*pos,Q)};
	U64 Kings[2]={piecesWHITE(*pos,K),piecesBLACK(*pos,K)};

	//the number of pieces for each side (except for kings)
	/*int numPawns[2]={popcnt(Pawns[WHITE]),popcnt(Pawns[BLACK])};
	int numKnights[2]={popcnt(Knights[WHITE]),popcnt(Knights[BLACK])};
	int numBishops[2]={popcnt(Bishops[WHITE]),popcnt(Bishops[BLACK])};
	int numRooks[2]={popcnt(Rooks[WHITE]),popcnt(Rooks[BLACK])};
	int numQueens[2]={popcnt(Queens[WHITE]),popcnt(Queens[BLACK])};*/
	int numPawns[2]={matsig_getP(pos->matsig[WHITE]),matsig_getP(pos->matsig[BLACK])};
	int numKnights[2]={matsig_getN(pos->matsig[WHITE]),matsig_getN(pos->matsig[BLACK])};
	int numBishops[2]={matsig_getB(pos->matsig[WHITE]),matsig_getB(pos->matsig[BLACK])};
	int numRooks[2]={matsig_getR(pos->matsig[WHITE]),matsig_getR(pos->matsig[BLACK])};
	int numQueens[2]={matsig_getQ(pos->matsig[WHITE]),matsig_getQ(pos->matsig[BLACK])};


	//simple material score of sliders and knights
	int pieceMaterial[2]=
	{
		Nval*numKnights[WHITE]+Bval*numBishops[WHITE]+Rval*numRooks[WHITE]+Qval*numQueens[WHITE],
		Nval*numKnights[BLACK]+Bval*numBishops[BLACK]+Rval*numRooks[BLACK]+Qval*numQueens[BLACK]
	};

	//attacks for each side
	U64 Pattacks[2];
	U64 PdoubleAttacks[2];
	U64 Battacks[2]={U64EMPTY,U64EMPTY};
	U64 Nattacks[2]={U64EMPTY,U64EMPTY};
	U64 NdoubleAttacks[2]={U64EMPTY,U64EMPTY};
	U64 BNattacks[2];
	U64 Rattacks[2]={U64EMPTY,U64EMPTY};
	U64 RdoubleAttacks[2]={U64EMPTY,U64EMPTY};
	U64 Qattacks[2]={U64EMPTY,U64EMPTY};
	U64 Kattacks[2]={Kmoves(pos->KingPos[WHITE]),Kmoves(pos->KingPos[BLACK])};
	U64 attacks[2]; //discluding King attacks
	U64 multiattacks[2]; //square attacked multiple times

	//safe squares for each piece type discluding King attacks
	/*U64 Psafe[2];
	U64 BNsafe[2]={U64EMPTY,U64EMPTY};
	U64 Rsafe[2]={U64EMPTY,U64EMPTY};
	U64 Qsafe[2]={U64EMPTY,U64EMPTY};*/

	//safe attacks for each side
	/*U64 BsafeAttacks[2];
	U64 NsafeAttacks[2];
	U64 RsafeAttacks[2];
	U64 QsafeAttacks[2];*/

	//pawns bitboards
	U64 immobilePawns[2];
	U64 Pdefendable[2]; //squares that can possibly be attacked by pawn attacks
	U64 Pdefendable_x2[2]; //squares that can be possible attacked twice by pawn attacks
	U64 Pblockable[2];  //squares that can possibly be occupied by pawn moves
	//U64 PdefendableSafe[2]; //squares that can safely be attacked by pawn attacks
	//U64 PblockableSafe[2];  //squares that can safely be occupied by pawn moves
	U64 doubledPawns[2];
	U64 isolatedPawns[2];
	U64 unprotectedPawns[2]; //Pawns that are not immediately protected by a pawn or side-to-side with a pawn
	U64 backwardPawns[2]; //A pawn that cannot advance beyond a enemy pawn formation with help from only it's own pawns
	U64 passedPawns[2];
	U64 halfpassedPawns[2];
	//Pawn Targets
	//Backward pawns+root pawns that are parallel to each other
	//+pawns that are not backward but are blocked from pawn defence
	//+pawns that are not defended at all
	//basically hard to defend pawns
	//U64 pawnTargets[2];

	/***********************************************************************/
	/*                            Data Generation                          */
	/***********************************************************************/

	/******************** Initialize Attack Bitboards ********************/

	//Pawn attacks
	Pattacks[WHITE]=((Pawns[WHITE]&C64(0x7F7F7F7F7F7F7F7F))<<9)|((Pawns[WHITE]&C64(0xFEFEFEFEFEFEFEFE))<<7);
	PdoubleAttacks[WHITE]=((Pawns[WHITE]&C64(0x7F7F7F7F7F7F7F7F))<<9)&((Pawns[WHITE]&C64(0xFEFEFEFEFEFEFEFE))<<7);
	Pattacks[BLACK]=((Pawns[BLACK]&C64(0x7F7F7F7F7F7F7F7F))>>7)|((Pawns[BLACK]&C64(0xFEFEFEFEFEFEFEFE))>>9);
	PdoubleAttacks[BLACK]=((Pawns[BLACK]&C64(0x7F7F7F7F7F7F7F7F))>>7)&((Pawns[BLACK]&C64(0xFEFEFEFEFEFEFEFE))>>9);
	{
		//Knight attacks
		U64 pieceboard=Knights[WHITE];
		while(pieceboard)
		{
			int index;
			GetBitAndClear(pieceboard,index);
			NdoubleAttacks[WHITE]|=Nattacks[WHITE]&Nmoves(index);
			Nattacks[WHITE]|=Nmoves(index);
		}
		pieceboard=Knights[BLACK];
		while(pieceboard)
		{
			int index;
			GetBitAndClear(pieceboard,index);
			NdoubleAttacks[BLACK]|=Nattacks[BLACK]&Nmoves(index);
			Nattacks[BLACK]|=Nmoves(index);
		}

		//Bishop attacks
		pieceboard=Bishops[WHITE];
		while(pieceboard)
		{
			int index;
			GetBitAndClear(pieceboard,index);
			Battacks[WHITE]|=Bmoves(index,pos->AllPieces);
		}
		pieceboard=Bishops[BLACK];
		while(pieceboard)
		{
			int index;
			GetBitAndClear(pieceboard,index);
			Battacks[BLACK]|=Bmoves(index,pos->AllPieces);
		}

		BNattacks[WHITE]=Battacks[WHITE]|Nattacks[WHITE];
		BNattacks[BLACK]=Battacks[BLACK]|Nattacks[BLACK];

		//Rook attacks
		pieceboard=Rooks[WHITE];
		while(pieceboard)
		{
			U64 temp;
			int index;
			GetBitAndClear(pieceboard,index);
			temp=Rmoves(index,pos->AllPieces);
			RdoubleAttacks[WHITE]|=Rattacks[WHITE]&temp;
			Rattacks[WHITE]|=temp;
		}
		pieceboard=Rooks[BLACK];
		while(pieceboard)
		{
			U64 temp;
			int index;
			GetBitAndClear(pieceboard,index);
			temp=Rmoves(index,pos->AllPieces);
			RdoubleAttacks[BLACK]|=Rattacks[BLACK]&temp;
			Rattacks[BLACK]|=temp;
		}

		//Queen attacks
		pieceboard=Queens[WHITE];
		while(pieceboard)
		{
			int index;
			GetBitAndClear(pieceboard,index);
			Qattacks[WHITE]|=Qmoves(index,pos->AllPieces);
		}
		pieceboard=Queens[BLACK];
		while(pieceboard)
		{
			int index;
			GetBitAndClear(pieceboard,index);
			Qattacks[BLACK]|=Qmoves(index,pos->AllPieces);
		}

		//King attacks are not considered true attacks for the purposes of kingsafety
		//Nevertheless - when doing flood fills you must take into account the king attacks
		attacks[WHITE]=Pattacks[WHITE]|Nattacks[WHITE]|Battacks[WHITE]|Rattacks[WHITE]|Qattacks[WHITE]|Kattacks[WHITE];
		attacks[BLACK]=Pattacks[BLACK]|Nattacks[BLACK]|Battacks[BLACK]|Rattacks[BLACK]|Qattacks[BLACK]|Kattacks[BLACK];

		//multi-atacks
		multiattacks[WHITE]=
			NdoubleAttacks[WHITE]|RdoubleAttacks[WHITE]|
			(Nattacks[WHITE]&Battacks[WHITE])|(Nattacks[WHITE]&Rattacks[WHITE])|(Nattacks[WHITE]&Qattacks[WHITE])|
			(Battacks[WHITE]&Rattacks[WHITE])|(Battacks[WHITE]&Qattacks[WHITE])|
			(Rattacks[WHITE]&Qattacks[WHITE])|
			(Kattacks[WHITE]&Nattacks[WHITE])|(Kattacks[WHITE]&Battacks[WHITE])|(Kattacks[WHITE]&Rattacks[WHITE])|
			(Kattacks[WHITE]&Qattacks[WHITE]);
		multiattacks[BLACK]=
			NdoubleAttacks[BLACK]|RdoubleAttacks[BLACK]|
			(Nattacks[BLACK]&Battacks[BLACK])|(Nattacks[BLACK]&Rattacks[BLACK])|(Nattacks[BLACK]&Qattacks[BLACK])|
			(Battacks[BLACK]&Rattacks[BLACK])|(Battacks[BLACK]&Qattacks[BLACK])|
			(Rattacks[BLACK]&Qattacks[BLACK])|
			(Kattacks[BLACK]&Nattacks[BLACK])|(Kattacks[BLACK]&Battacks[BLACK])|(Kattacks[BLACK]&Rattacks[BLACK])|
			(Kattacks[BLACK]&Qattacks[BLACK]);
	}

	/******************** Initialize Safe Squares ********************/
	/*{
		U64 unsafe[2]={~attacks[WHITE]&attacks[BLACK],~attacks[BLACK]&attacks[WHITE]};
		Psafe[WHITE]=~(unsafe[WHITE]|pos->AllPieces);
		Psafe[BLACK]=~(unsafe[BLACK]|pos->AllPieces);
		unsafe[WHITE]|=Pattacks[BLACK];
		unsafe[BLACK]|=Pattacks[WHITE];
		unsafe[WHITE]|=(attacks[BLACK]&~multiattacks[WHITE])|multiattacks[BLACK];
		unsafe[BLACK]|=(attacks[WHITE]&~multiattacks[BLACK])|multiattacks[WHITE];
		BNsafe[WHITE]=~unsafe[WHITE];
		BNsafe[BLACK]=~unsafe[BLACK];
		unsafe[WHITE]|=Nattacks[BLACK]|Battacks[BLACK];
		unsafe[BLACK]|=Nattacks[WHITE]|Battacks[WHITE];
		Rsafe[WHITE]=~unsafe[WHITE];
		Rsafe[BLACK]=~unsafe[BLACK];
		unsafe[WHITE]|=Rattacks[BLACK];
		unsafe[BLACK]|=Rattacks[WHITE];
		Qsafe[WHITE]=~unsafe[WHITE];
		Qsafe[BLACK]=~unsafe[BLACK];
	}

	/******************** Initialize Safe Attacks ********************/
	/*BsafeAttacks[WHITE]=Battacks[WHITE]&BNsafe[WHITE];
	NsafeAttacks[WHITE]=Nattacks[WHITE]&BNsafe[WHITE];
	RsafeAttacks[WHITE]=Rattacks[WHITE]&Rsafe[WHITE];
	QsafeAttacks[WHITE]=Qattacks[WHITE]&Qsafe[WHITE];
	BsafeAttacks[BLACK]=Battacks[BLACK]&BNsafe[BLACK];
	NsafeAttacks[BLACK]=Nattacks[BLACK]&BNsafe[BLACK];
	RsafeAttacks[BLACK]=Rattacks[BLACK]&Rsafe[BLACK];
	QsafeAttacks[BLACK]=Qattacks[BLACK]&Qsafe[BLACK];

	/******************** Initialize Pawn Structure Bitboards ********************/
	immobilePawns[WHITE]=Pawns[WHITE]&((pos->PiecesSide[BLACK]|Pawns[WHITE])>>8);
	immobilePawns[BLACK]=Pawns[BLACK]&((pos->PiecesSide[WHITE]|Pawns[BLACK])<<8);

	Pblockable[WHITE] = fillUp2(Pawns[WHITE]);
	Pblockable[BLACK] = fillDown2(Pawns[BLACK]);

	/*PblockableSafe[WHITE]=fillUpOccluded(Pawns[WHITE],Psafe[WHITE]);
	PblockableSafe[BLACK]=fillDownOccluded(Pawns[BLACK],Psafe[BLACK]);*/

	Pdefendable[WHITE]=((Pblockable[WHITE]&C64(0x7F7F7F7F7F7F7F7F))<<9)|((Pblockable[WHITE]&C64(0xFEFEFEFEFEFEFEFE))<<7);
	Pdefendable[BLACK]=((Pblockable[BLACK]&C64(0x7F7F7F7F7F7F7F7F))>>7)|((Pblockable[BLACK]&C64(0xFEFEFEFEFEFEFEFE))>>9);

	Pdefendable_x2[WHITE]=((Pblockable[WHITE]&C64(0x7F7F7F7F7F7F7F7F))<<9)&((Pblockable[WHITE]&C64(0xFEFEFEFEFEFEFEFE))<<7);
	Pdefendable_x2[BLACK]=((Pblockable[BLACK]&C64(0x7F7F7F7F7F7F7F7F))>>7)&((Pblockable[BLACK]&C64(0xFEFEFEFEFEFEFEFE))>>9);

	/*PdefendableSafe[WHITE]=((PblockableSafe[WHITE]&C64(0x7F7F7F7F7F7F7F7F))<<9)|((PblockableSafe[WHITE]&C64(0xFEFEFEFEFEFEFEFE))<<7);
	PdefendableSafe[BLACK]=((PblockableSafe[BLACK]&C64(0x7F7F7F7F7F7F7F7F))>>7)|((PblockableSafe[BLACK]&C64(0xFEFEFEFEFEFEFEFE))>>9);*/

	doubledPawns[WHITE]=Pawns[WHITE] & fillDown(Pawns[WHITE]);
	doubledPawns[BLACK]=Pawns[BLACK] & fillUp(Pawns[BLACK]);

	passedPawns[WHITE]=Pawns[WHITE] & ~doubledPawns[WHITE] & ~(Pdefendable[BLACK]|Pblockable[BLACK]);
	passedPawns[BLACK]=Pawns[BLACK] & ~doubledPawns[BLACK] & ~(Pdefendable[WHITE]|Pblockable[WHITE]);

	unprotectedPawns[WHITE]=Pawns[WHITE]&~fillUp2(Pattacks[WHITE]>>8 & ~immobilePawns[WHITE]);
	unprotectedPawns[BLACK]=Pawns[BLACK]&~fillDown2(Pattacks[BLACK]<<8 & ~immobilePawns[BLACK]);

	backwardPawns[WHITE]=Pawns[WHITE] &
		fillDown((Pattacks[BLACK]&~Pdefendable[WHITE]) | (PdoubleAttacks[BLACK]&~Pdefendable_x2[WHITE]));
	backwardPawns[BLACK]=Pawns[BLACK] &
		fillUp((Pattacks[WHITE]&~Pdefendable[BLACK]) | (PdoubleAttacks[WHITE]&~Pdefendable_x2[BLACK]));

	isolatedPawns[WHITE]=backwardPawns[WHITE]&~fillDown2(Pattacks[WHITE]>>8);
	isolatedPawns[BLACK]=backwardPawns[BLACK]&~fillUp2(Pattacks[BLACK]<<8);

	halfpassedPawns[WHITE]=~backwardPawns[WHITE]&(~passedPawns[WHITE])&(
		(Pawns[WHITE] & ~fillDown2(Pawns[BLACK]|((Pawns[BLACK]&C64(0x7F7F7F7F7F7F7F7F))>>7)))|
		(Pawns[WHITE] & ~fillDown2(Pawns[BLACK]|((Pawns[BLACK]&C64(0xFEFEFEFEFEFEFEFE))>>9))));
	halfpassedPawns[BLACK]=~backwardPawns[BLACK]&(~passedPawns[BLACK])&(
		(Pawns[BLACK] & ~fillUp2(Pawns[WHITE]|((Pawns[WHITE]&C64(0x7F7F7F7F7F7F7F7F))<<9)))|
		(Pawns[BLACK] & ~fillUp2(Pawns[WHITE]|((Pawns[WHITE]&C64(0xFEFEFEFEFEFEFEFE))<<7))));


	/*pawnTargets[WHITE]=Pawns[BLACK]&(~attacks[BLACK] |
		~fillDownOccluded(Pattacks[BLACK]<<8 & ~immobilePawns[BLACK],~(Pattacks[WHITE]>>8)));
	pawnTargets[BLACK]=Pawns[WHITE]&(~attacks[WHITE] |
		~fillUpOccluded(Pattacks[WHITE]>>8 & ~immobilePawns[WHITE],~(Pattacks[BLACK]<<8)));*/


	/***********************************************************************/
	/*                            Data analysis                            */
	/***********************************************************************/

	//Listed in order of decreasing importance for lazy eval

	/******************** Material and Piece Square Evaluation ********************/
	#ifdef MATERIAL_EVAL
	{
		//Pawn Material Eval is done in Pawn Eval
		U64 WhitePawns = Pawns[WHITE];
		U64 BlackPawns = Pawns[BLACK];

		//Pawn Material Eval done by Piece Square
		while(WhitePawns)
		{
			int index;
			GetBitAndClear(WhitePawns,index);
			op[WHITE]+=P_op_PST[index];
			eg[WHITE]+=P_eg_PST[index];

			//Distance of own king to pawn should be scaled by own material
			//Distance of opponent king to pawn should be scaled by opponent material
			#ifdef KING_DISTANCE_EVAL
			eg[BLACK]+=distance(index,pos->KingPos[WHITE])*RankBonus2[RANK(index)]*Kdist_side/256;
			eg[WHITE]+=distance(index,pos->KingPos[BLACK])*RankBonus2[RANK(index)]*Kdist_xside/256;
			#endif
		}
		while(BlackPawns)
		{
			int index;
			GetBitAndClear(BlackPawns,index);
			op[BLACK]+=P_op_PST[FLIPV(index)];
			eg[BLACK]+=P_eg_PST[FLIPV(index)];

			//Distance of own king to pawn should be scaled by own material
			//Distance of opponent king to pawn should be scaled by opponent material
			#ifdef KING_DISTANCE_EVAL
			eg[WHITE]+=distance(index,pos->KingPos[BLACK])*RankBonus2[RANK(FLIPV(index))]*Kdist_side/256;
			eg[BLACK]+=distance(index,pos->KingPos[WHITE])*RankBonus2[RANK(FLIPV(index))]*Kdist_xside/256;
			#endif
		}

		DEBUG_EVAL_PRINT("Pawn Material Eval and King Distance")

		//King Piece Square Eval
		op[WHITE]+=K_op_PST[pos->KingPos[WHITE]];
		eg[WHITE]+=K_eg_PST[pos->KingPos[WHITE]];
		op[BLACK]+=K_op_PST[FLIPV(pos->KingPos[BLACK])];
		eg[BLACK]+=K_eg_PST[FLIPV(pos->KingPos[BLACK])];

		DEBUG_EVAL_PRINT("King Piece Square Eval")

		//Regular Material Eval
		op[WHITE]+=
			N_op*(numKnights[WHITE]) +
			B_op*(numBishops[WHITE]) +
			R_op*(numRooks[WHITE])   +
			Q_op*(numQueens[WHITE]);
		eg[WHITE]+=
			N_eg*(numKnights[WHITE]) +
			B_eg*(numBishops[WHITE]) +
			R_eg*(numRooks[WHITE])   +
			Q_eg*(numQueens[WHITE]);
		op[BLACK]+=
			N_op*(numKnights[BLACK]) +
			B_op*(numBishops[BLACK]) +
			R_op*(numRooks[BLACK])   +
			Q_op*(numQueens[BLACK]);
		eg[BLACK]+=
			N_eg*(numKnights[BLACK]) +
			B_eg*(numBishops[BLACK]) +
			R_eg*(numRooks[BLACK])   +
			Q_eg*(numQueens[BLACK]);

		DEBUG_EVAL_PRINT("Piece Material Eval")
	}
	#endif


	/******************** Pawn Evaluvation ********************/

	#ifdef PAWN_EVAL
	{
		U64 doubledWHITE=doubledPawns[WHITE];
		U64 doubledBLACK=doubledPawns[BLACK];


		U64 isolatedWHITE=isolatedPawns[WHITE];
		U64 isolatedBLACK=isolatedPawns[BLACK];

		U64 unprotectedWHITE = unprotectedPawns[WHITE];
		U64 unprotectedBLACK = unprotectedPawns[BLACK];

		/*U64 backwardWHITE=backwardPawns[WHITE]&~(isolatedPawns[WHITE]|doubledPawns[WHITE]);
		U64 backwardBLACK=backwardPawns[BLACK]&~(isolatedPawns[BLACK]|doubledPawns[BLACK]);*/

		U64 passedWHITE=passedPawns[WHITE];
		U64 passedBLACK=passedPawns[BLACK];

		U64 halfpassedWHITE=halfpassedPawns[WHITE];
		U64 halfpassedBLACK=halfpassedPawns[BLACK];

		while(doubledWHITE)
		{
			int sq;
			GetBitAndClear(doubledWHITE,sq);
			op[WHITE]-=doubledPenalty_op;
			eg[WHITE]-=doubledPenalty_eg;
		}
		while(doubledBLACK)
		{
			int sq;
			GetBitAndClear(doubledBLACK,sq);
			op[BLACK]-=doubledPenalty_op;
			eg[BLACK]-=doubledPenalty_eg;
		}
		DEBUG_EVAL_PRINT("Doubled Pawn Eval")

		//unprotected pawns
		while(unprotectedWHITE)
		{
			int sq;
			GetBitAndClear(unprotectedWHITE,sq);
			op[WHITE]-=unprotectedPenalty_op;
			eg[WHITE]-=unprotectedPenalty_eg;
		}
		while(unprotectedBLACK)
		{
			int sq;
			GetBitAndClear(unprotectedBLACK,sq);
			op[BLACK]-=unprotectedPenalty_op;
			eg[BLACK]-=unprotectedPenalty_eg;
		}
		DEBUG_EVAL_PRINT("Unprotected Pawn Eval")

		//backward pawns
		/*while(backwardWHITE)
		{
			int sq;
			GetBitAndClear(backwardWHITE,sq);
			op[WHITE]-=backwardPenalty_op;
			eg[WHITE]-=backwardPenalty_eg;
		}
		while(backwardBLACK)
		{
			int sq;
			GetBitAndClear(backwardBLACK,sq);
			op[BLACK]-=backwardPenalty_op;
			eg[BLACK]-=backwardPenalty_eg;
		}
		DEBUG_EVAL_PRINT("Backward Pawn Eval")*/

		//isolated pawns
		while(isolatedWHITE)
		{
			int sq;
			GetBitAndClear(isolatedWHITE,sq);
			op[WHITE]-=isolatedPenalty_op;
			eg[WHITE]-=isolatedPenalty_eg;
			//Bonus for enemy king too far
			//Bonus for our king close to it
		}
		while(isolatedBLACK)
		{
			int sq;
			GetBitAndClear(isolatedBLACK,sq);
			op[BLACK]-=isolatedPenalty_op;
			eg[BLACK]-=isolatedPenalty_eg;
			//Bonus for enemy king too far
			//Bonus for our king close to it
		}
		DEBUG_EVAL_PRINT("Isolated Pawn Eval")

		//Passed pawns

		while(passedWHITE)
		{
			int sq;
			GetBitAndClear(passedWHITE,sq);
			op[WHITE]+=passerBonus_op+passerRBonus_op*RankBonus[RANK(sq)]/256;
			eg[WHITE]+=passerBonus_eg+passerRBonus_eg*RankBonus[RANK(sq)]/256;

			if(passedPawns[WHITE]&((FILE(sq)!=FILEA?FILEBB(FILE(sq)-1):0) | (FILE(sq)!=FILEH?FILEBB(FILE(sq)+1):0)))
			{
				op[WHITE]+=connectedPasser_Bonus_op;
				eg[WHITE]+=connectedPasser_Bonus_eg;
			}

			#ifdef KING_DISTANCE_EVAL
			eg[BLACK]+=distance(sq+8,pos->KingPos[WHITE])*RankBonus2[RANK(sq)]*Kdist_passer_side/256;
			eg[WHITE]+=distance(sq+8,pos->KingPos[BLACK])*RankBonus2[RANK(sq)]*Kdist_passer_xside/256;
			#endif
		}
		while(passedBLACK)
		{
			int sq;
			GetBitAndClear(passedBLACK,sq);
			op[BLACK]+=passerBonus_op+passerRBonus_op*RankBonus[RANK(FLIPV(sq))]/256;
			eg[BLACK]+=passerBonus_eg+passerRBonus_eg*RankBonus[RANK(FLIPV(sq))]/256;

			if(passedPawns[BLACK]&((FILE(sq)!=FILEA?FILEBB(FILE(sq)-1):0) | (FILE(sq)!=FILEH?FILEBB(FILE(sq)+1):0)))
			{
				op[BLACK]+=connectedPasser_Bonus_op;
				eg[BLACK]+=connectedPasser_Bonus_eg;
			}

			/*if(pos->AllPieces^(Kings[WHITE]|Kings[BLACK]))
			{
				if(pos->side==WHITE)
				{
					if(distance(KingPos[WHITE]))
				}

			}*/

			#ifdef KING_DISTANCE_EVAL
			eg[WHITE]+=distance(sq-8,pos->KingPos[BLACK])*RankBonus2[RANK(FLIPV(sq))]*Kdist_passer_side/256;
			eg[BLACK]+=distance(sq-8,pos->KingPos[WHITE])*RankBonus2[RANK(FLIPV(sq))]*Kdist_passer_xside/256;
			#endif
		}
		DEBUG_EVAL_PRINT("Passed Pawn Eval")

		//half-passed pawns that are defendable

		while(halfpassedWHITE)
		{
			int sq;
			GetBitAndClear(halfpassedWHITE,sq);
			op[WHITE]+=halfPasserBonus_op+halfPasserRBonus_op*RankBonus[RANK(sq)]/256;
			eg[WHITE]+=halfPasserBonus_eg+halfPasserRBonus_eg*RankBonus[RANK(sq)]/256;

			#ifdef KING_DISTANCE_EVAL
			eg[BLACK]+=distance(sq+8,pos->KingPos[WHITE])*RankBonus2[RANK(sq)]*Kdist_halfpasser_side/256;
			eg[WHITE]+=distance(sq+8,pos->KingPos[BLACK])*RankBonus2[RANK(sq)]*Kdist_halfpasser_xside/256;
			#endif
			//Bonus for enemy king too far
			//Bonus for our king close to it
		}
		while(halfpassedBLACK)
		{
			int sq;
			GetBitAndClear(halfpassedBLACK,sq);
			op[BLACK]+=halfPasserBonus_op+halfPasserRBonus_op*RankBonus[RANK(FLIPV(sq))]/256;
			eg[BLACK]+=halfPasserBonus_eg+halfPasserRBonus_eg*RankBonus[RANK(FLIPV(sq))]/256;

			#ifdef KING_DISTANCE_EVAL
			eg[WHITE]+=distance(sq-8,pos->KingPos[BLACK])*RankBonus2[RANK(FLIPV(sq))]*Kdist_halfpasser_side/256;
			eg[BLACK]+=distance(sq-8,pos->KingPos[WHITE])*RankBonus2[RANK(FLIPV(sq))]*Kdist_halfpasser_xside/256;
			#endif
			//Bonus for enemy king too far
			//Bonus for our king close to it
		}
		DEBUG_EVAL_PRINT("Half-passed Pawn Eval")
		//DEBUG_EVAL_PRINT("Pawn Eval")
	}
	#endif

	/******************** Mobility Evaluation ********************/
	#ifdef MOBILITY_EVAL
	{
		int mobility[2];
		mobility[WHITE] =
			MobB_op*popcnt(Battacks[WHITE]) +
			MobN_op*popcnt(Nattacks[WHITE]) +
			MobR_op*popcnt(Rattacks[WHITE]) +
			MobQ_op*popcnt(Qattacks[WHITE]) ;

		mobility[BLACK] =
			MobB_op*popcnt(Battacks[BLACK]) +
			MobN_op*popcnt(Nattacks[BLACK]) +
			MobR_op*popcnt(Rattacks[BLACK]) +
			MobQ_op*popcnt(Qattacks[BLACK]) ;

		score += mobility[pos->side] - mobility[pos->xside];

		DEBUG_EVAL_PRINT("Mobility Eval")
	}
	#endif

	/******************** Side to move bonus ********************/
	#ifdef STM_BONUS
	op[pos->side]+=STM_BONUS;
	#endif STM_BONUS

	/******************** Kingsafety Eval ********************/



	/******************** Compile Scores ********************/
	{
		int compiled_score[2] = {
			SCALE_SCORE(op[WHITE],eg[WHITE],pieceMaterial[BLACK]),
			SCALE_SCORE(op[BLACK],eg[BLACK],pieceMaterial[WHITE])
		};
		score += compiled_score[pos->side] - compiled_score[pos->xside];
	}

	/*if(score>=RECOG_WIN)
		score = RECOG_WIN-1;
	else if(score<=-RECOG_WIN)
		score = -RECOG_WIN+1;*/

	return score;
}
#endif

//checks if the eval for the current position is symmetrical
//this should always return true unless there is a bug
bool symmetryCheck(const board* b)
{
	board flipb = mirror(b);
	if(eval(&flipb,-INF,INF)!=eval(b,-INF,INF))
	{
		/*print("Symmetry Check Failed!\n");
		printBoard(b);
		fen(b);
		print("Eval = %d\n",eval(b,-INF,INF));
		printBoard(&flipb);
		fen(&flipb);
		print("Eval = %d\n",eval(&flipb,-INF,INF));
		system("Pause");*/
		return false;
	}
	return true;
}

int materialWHITE(const board* pos)
{
	return   Pval*(popcnt(piecesWHITE(*pos,P)) - popcnt(piecesBLACK(*pos,P)))
			+Nval*(popcnt(piecesWHITE(*pos,N)) - popcnt(piecesBLACK(*pos,N)))
			+Rval*(popcnt(piecesWHITE(*pos,R)) - popcnt(piecesBLACK(*pos,R)))
			+Bval*(popcnt(piecesWHITE(*pos,B)) - popcnt(piecesBLACK(*pos,B)))
			+Qval*(popcnt(piecesWHITE(*pos,Q)) - popcnt(piecesBLACK(*pos,Q)));
}

int materialBLACK(const board* pos)
{
	return  -materialWHITE(pos);
}

int materialSide(const board* pos)
{
	if(pos->side) //Black
		return materialBLACK(pos);
	else
		return materialWHITE(pos);
}

int materialXSide(const board* pos)
{
	return -materialSide(pos);
}

/*******************************************************************************
*                                  ANDRES EVAL                                 *
********************************************************************************/




#ifdef ANDRES_EVAL

const int Square_Color_PST [64]= {

              0,1,0,1,0,1,0,1,
              1,0,1,0,1,0,1,0,
              0,1,0,1,0,1,0,1,
              1,0,1,0,1,0,1,0,
              0,1,0,1,0,1,0,1,
              1,0,1,0,1,0,1,0,
              0,1,0,1,0,1,0,1,
			  1,0,1,0,1,0,1,0};

const int Pawn_PST[2][64] =

{{ 0,  0,  0,  0,  0,  0,  0,  0,
 -17, -5,  1,  8,  8,  1, -5,-17,
 -18, -6,  0,  7,  7,  0, -6,-18,
 -19, -7, -1,  6,  6, -1, -7,-19,
 -21, -9, -3,  4,  4, -3, -9,-21,
 -22,-10, -4,  3,  3, -4,-10,-22,
 -23,-11, -5,  2,  2, -5,-11,-23,
   0,  0,  0,  0,  0,  0,  0,  0},

 { 0,  0,  0,  0,  0,  0,  0,  0,
  -2, -4, -6, -8, -8, -6, -4, -2,
  -4, -6, -8,-10,-10, -8, -6, -4,
  -5, -7, -9,-11,-11, -9, -7, -5,
  -6, -8,-10,-12,-12,-10, -8, -6,
  -7, -9,-11,-13,-13,-11, -9, -7,
  -7, -9,-11,-13,-13,-11, -9, -7,
   0,  0,  0,  0,  0,  0,  0,  0}};


const int Knight_PST[2][64]=

{{
-135, -25, -15, -10, -10, -15, -25,-135,
 -20, -10,   0,   5,   5,   0, -10, -20,
  -5,   5,  15,  20,  20,  15,   5,  -5,
  -5,   5,  15,  20,  20,  15,   5,  -5,
 -10,   0,  10,  15,  15,  10,   0, -10,
 -20, -10,   0,   5,   5,   0, -10, -20,
 -35, -25, -15, -10, -10, -15, -25, -35,
 -50, -40, -30, -25, -25, -30, -40, -50
},



{-40, -30, -20, -15, -15, -20, -30, -40,
 -30, -20, -10,  -5,  -5, -10, -20, -30,
 -20, -10,   0,   5,   5,   0, -10, -20,
 -15,  -5,   5,  10,  10,   5,  -5, -15,
 -15,  -5,   5,  10,  10,   5,  -5, -15,
 -20, -10,   0,   5,   5,   0, -10, -20,
 -30, -20, -10,  -5,  -5, -10, -20, -30,
 -40, -30, -20, -15, -15, -20, -30, -40}};

const int Bishop_PST[2][64] =

{{-8,  -8,  -6,  -4,  -4,  -6,  -8,  -8,
  -8,   0,  -2,   0,   0,  -2,   0,  -8,
  -6,  -2,   4,   2,   2,   4,  -2,  -6,
  -4,   0,   2,   8,   8,   2,   0,  -4,
  -4,   0,   2,   8,   8,   2,   0,  -4,
  -6,  -2,   4,   2,   2,   4,  -2,  -6,
  -8,   0,  -2,   0,   0,  -2,   0,  -8,
 -18, - 6,  -9,  -6,  -6,  -9, - 6, -18},

{-18, -12,  -9,  -6,  -6,  -9, -12, -18,
 -12,  -6,  -3,   0,   0,  -3,  -6, -12,
  -9,  -3,   0,   3,   3,   0,  -3,  -9,
  -6,   0,   3,   6,   6,   3,   0,  -6,
  -6,   0,   3,   6,   6,   3,   0,  -6,
  -9,  -3,   0,   3,   3,   0,  -3,  -9,
 -12,  -6,  -3,   0,   0,  -3,  -6, -12,
 -18, -12,  -9,  -6,  -6,  -9, -12, -18}};


const int King_PST[2][64]=

{{-40, -40, -40, -40, -40, -40, -40, -40,
  -40, -40, -40, -40, -40, -40, -40, -40,
  -40, -40, -40, -40, -40, -40, -40, -40,
  -40, -40, -40, -40, -40, -40, -40, -40,
  -40, -40, -40, -40, -40, -40, -40, -40,
  -40, -40, -40, -40, -40, -40, -40, -40,
   10,  10, -20, -35, -35, -25,  10,  10,
    0,  20,   5, -35, -30, -10,  20,   0},

{-72, -48, -36, -24, -24, -36, -48, -72,
 -48, -24, -12,   0,   0, -12, -24, -48,
 -36, -12,   0,  12,  12,   0, -12, -36,
 -24,   0,  12,  24,  24,  12,   0, -24,
 -24,   0,  12,  24,  24,  12,   0, -24,
 -36, -12,   0,  12,  12,   0, -12, -36,
 -48, -24, -12,   0,   0, -12, -24, -48,
 -72, -48, -36, -24, -24, -36, -48, -72}};

const int Rook_PST[64] =

{ 10,  12,  14,  16,  16,  14,  12,  10,
  20,  22,  24,  26,  26,  24,  22,  20,
  -6,  -3,   0,   3,   3,   0,  -3,  -6,
  -6,  -3,   0,   3,   3,   0,  -3,  -6,
  -6,  -3,   0,   3,   3,   0,  -3,  -6,
  -6,  -3,   0,   3,   3,   0,  -3,  -6,
  -6,  -3,   0,   3,   3,   0,  -3,  -6,
  -6,  -3,   3,   6,   6,   3,  -3,  -6};

const int Backward_Pawn_PST [64]=
              { 0,  0,  0,  0,  0,  0,  0,  0,
                6,  8, 10, 15, 15, 10,  8,  6,
                6,  8, 10, 15, 15, 10,  8,  6,
                6,  8, 10, 20, 20, 10,  8,  6,
                6,  8, 15, 20, 20, 15,  8,  6,
                4,  6, 15, 20, 20, 15,  6,  4,
                2,  4,  8, 10, 10,  8,  4,  2,
			    0,  0,  0,  0,  0,  0,  0,  0};

#ifdef KING_SAFETY_EVAL

const int KSCounterTable[64]=

             {4,4,4,4,4,4,4,4,
              4,4,4,4,4,4,4,4,
              4,4,4,4,4,4,4,4,
              4,4,4,4,4,4,4,4,
              4,4,4,4,4,4,4,4,
              4,3,3,3,3,3,3,4,
              0,0,1,1,1,1,0,0,
			  0,0,0,0,0,0,0,0};

const int KSTable[50] =

              {0,  2,  3,  6, 12, 18, 25, 37, 50, 75,
              100,125,150,175,200,225,250,275,300,325,
              350,375,400,425,450,475,500,525,550,575,
              600,600,600,600,600,600,600,600,600,600,
			  600,600,600,600,600,600,600,600,600,600};

/*
 0, 1, 2, 3, 4, 5, 6, 7
 8, 9,10,11,12,13,14,15,
16,17,18,19,20,21,22,23,
24,25,26,27,28,29,30,31,  <------ BUZZ BOARD REPRESENTATION. 0 is H1, 63 is a8
32,33,34,35,36,37,38,39,
40,41,42,43,44,45,46,47,
48,49,50,51,52,53,54,55,
56,57,58,59,60,61,62,63;
*/

// King barrier are three squares beyond the squares inmediately around the king
//
//   X X X
//   o o o    <-- White King barrier squares as "X"
//   o K o
//
// This is a precalculated table. Only takes in account three first rows because
// a more advanced king has a lot of more dangers than those 3 squares :-)


const int WKingBarrier [24][3] =

   {{A3,B3,C3},{A3,B3,C3},{B3,C3,D3},{C3,D3,E3},{D3,E3,F3},{E3,F3,G3},{F3,G3,H3},{F3,G3,H3},
	{A4,B4,C4},{A4,B4,C4},{B4,C4,D4},{C4,D4,E4},{D4,E4,F4},{E4,F4,G4},{F4,G4,H4},{F4,G4,H4},
	{A5,B5,C5},{A5,B5,C5},{B5,C5,D5},{C5,D5,E5},{D5,E5,F5},{E5,F5,G5},{F5,G5,H5},{F5,G5,H5}};

const int BKingBarrier [24][3] =

   {{A6,B6,C6},{A6,B6,C6},{B6,C6,D6},{C6,D6,E6},{D6,E6,F6},{E6,F6,G6},{F6,G6,H6},{F6,G6,H6},
	{A5,B5,C5},{A5,B5,C5},{B5,C5,D5},{C5,D5,E5},{D5,E5,F5},{E5,F5,G5},{F5,G5,H5},{F5,G5,H5},
	{A4,B4,C4},{A4,B4,C4},{B4,C4,D4},{C4,D4,E4},{D4,E4,F4},{E4,F4,G4},{F4,G4,H4},{F4,G4,H4}};


const int Pawn_Shelter_Bonus [10] = {-75,-50,-25,0,0,5,10,15,20,25};
#endif

//const int PST_Weight [2] = {100,100};
const int PST_Weight [2] = {75,75};
const int Mobility_Weight [2] = {100,100};
const int Pawn_Weight [2] = {100,100};
//const int Passer_Pawn_Weight [2] = {75,100};
const int Passer_Pawn_Weight [2] = {50,75};
//const int King_Safety_Weight [2] = {100,75};
//const int King_Safety_Weight [2] = {75,50};
const int King_Safety_Weight [2] = {60,40};

//const int Lazy_1_Margin [2] = {300,300};
//const int Lazy_2_Margin [2] = {150,300}; //150,300
//const int Lazy_3_Margin [2] = {100, 50}; //150,50

const int Lazy_1_Margin [2] = {400,400};
const int Lazy_2_Margin [2] = {400,400}; //150,300

const int Lazy_3_Margin [2] = {100, 100}; //150,50

int fdist_array[64][64];


//#include "evalcache.h"
//load Dll
//typedef int (__stdcall *importFunction)(int side, int alpha, int beta, int pieceList[2][16], int pieces[64], int color[64], int passed[2][8], int halfpassed[2][8], int doubled[2][8], int isolated[2][8], int backward[2][8]);
//typedef int (__stdcall *importFunction)(int side, int alpha, int beta, int pieceList[2][16], int pieces[64], int color[64]);
//typedef int (__stdcall *importFunction)(const board* pos, int alpha, int beta);
//typedef int (__stdcall *importInitFunction)(char* path, char* password, int cache_size);

char degbb[4096]="";
int dcache=32;
bool egbb_is_loaded=false;
static PPROBE_EGBB probe_egbb;
/*
void init_AndresEval()
{
	HINSTANCE hinstLib = NULL;
	hinstLib=LoadLibrary("dirty.dll");
	if(hinstLib == NULL)
	{
		printf("Error could not load eval dll\n");
		exit(EXIT_FAILURE);
	}
	dirtyeval = (importFunction)GetProcAddress(hinstLib, "DirtyEval");
	InitDirtyDll = (importInitFunction)GetProcAddress(hinstLib, "InitDirtyDll");

	if(dirtyeval==NULL)
	{
		printf("Error could not get function pointer for eval from dirty.dll\n");
		exit(EXIT_FAILURE);
	}
	if(InitDirtyDll==NULL)
	{
		printf("Error could not get function pointer for init from dirty.dll\n");
		exit(EXIT_FAILURE);
	}
	InitDirtyDll(degbb,"DIRTY DOZEN INC.",dcache);
	//initEvalCache();
	//resetEvalCache();
}

*/
///////////////////////////////////////////////////
//////////// AUXILIAR FUNCTIONS
///////////////////////////////////////////////////
int frow(int sq)
{
return sq/8;
}

int fcol(int sq)
{
return sq & 7;
}


int fdistance(int sq1, int sq2)
{
return max(abs(fcol(sq2)-fcol(sq1)),abs(frow(sq2)-frow(sq1)));
}

void init_bitbases()
{
	egbb_is_loaded=false;
	egbb_is_loaded=LoadEgbbLibrary(degbb,dcache,&probe_egbb);
}


/*
Probe:
TSCP uses A8 = 0, H1 = 63 but we use A1 = 0, H8 = 63
board representation and then probe bitbase.
*/


//better move this to the eval or else within dirty
//just let egbb.c set the right pointer into the dll
int probe_bitbases(const board* pos)

{

	int sq, piece[3],square[3],count;
	int player = pos->side;

    int w_ksq = pos->KingPos[WHITE];
	int b_ksq = pos->KingPos[BLACK];


	U64 WhitePawns=piecesWHITE(*pos,P);
    U64 BlackPawns=piecesBLACK(*pos,P);
    U64 WhiteKnights=piecesWHITE(*pos,N);
    U64 BlackKnights=piecesBLACK(*pos,N);
    U64 WhiteBishops=piecesWHITE(*pos,B);
    U64 BlackBishops=piecesBLACK(*pos,B);
    U64 WhiteRooks=piecesWHITE(*pos,R);
    U64 BlackRooks=piecesBLACK(*pos,R);
	U64 WhiteQueens=piecesWHITE(*pos,Q);
    U64 BlackQueens=piecesBLACK(*pos,Q);


	count = 0;
	piece[0] = 0;
	piece[1] = 0;
	piece[2] = 0;
	square[0] = 0;
	square[1] = 0;
	square[2] = 0;

/*
	{
	    #define ADD_PIECE(type)
		 {\
	    piece[count] = type;\
	    square[count] = sq;\
        count++;\
		 };
	 }
*/

     while (WhitePawns)
	 {
	 GetBitAndClear(WhitePawns,sq);
    // sq=FLIP(sq);
	 piece[count] = _WPAWN;
  	 square[count] = sq;
     count++;
	 }

	 while (BlackPawns)
	 {
	 GetBitAndClear(BlackPawns,sq);
	 piece[count] = _BPAWN;
  	 square[count] = sq;
     count++;
	 }

	 while (WhiteKnights)
	 {
	 GetBitAndClear(WhiteKnights,sq);
	 piece[count] = _WKNIGHT;
  	 square[count] = sq;
     count++;
	 }

     while (BlackKnights)
	 {
	 GetBitAndClear(BlackKnights,sq);
	 piece[count] = _BKNIGHT;
  	 square[count] = sq;
     count++;
	 }
	 while (WhiteBishops)
	 {
	 GetBitAndClear(WhiteBishops,sq);
	 piece[count] = _WBISHOP;
  	 square[count] = sq;
     count++;
	 }

     while (BlackBishops)
	 {
	 GetBitAndClear(BlackBishops,sq);
	 piece[count] = _BBISHOP;
  	 square[count] = sq;
     count++;
	 }
     while (WhiteRooks)
	 {
	 GetBitAndClear(WhiteRooks,sq);
	 piece[count] = _WROOK;
  	 square[count] = sq;
     count++;
	 }

     while (BlackRooks)
	 {
	 GetBitAndClear(BlackRooks,sq);
	 piece[count] = _BROOK;
  	 square[count] = sq;
     count++;
	 }

    while (WhiteQueens)
	 {
	 GetBitAndClear(WhiteQueens,sq);
	 piece[count] = _WQUEEN;
  	 square[count] = sq;
     count++;
	 }

     while (BlackQueens)
	 {
	 GetBitAndClear(BlackQueens,sq);
	 piece[count] = _BQUEEN;
  	 square[count] = sq;
     count++;
	 }

	return probe_egbb(player, w_ksq, b_ksq,
		piece[0],square[0],piece[1],square[1],piece[2],square[2]);
}


///////////////////////////////////////////////////
//////////// MAIN EVAL FUNCTION
///////////////////////////////////////////////////

int AndresEval(const board* pos, int alpha, int beta)
{
	// Different scores
	//int piece[64],color[64];
	int Score[2]={0,0};  // Total score for each side
	int	Material_Score[2]={0,0};
	int PST_Score[2]={0,0};
	int Pawn_Score[2]={0,0};
	int Passer_Score[2]={0,0};
	int mobility[2]={0,0};
	int KingSafety[2]={0,0};
	//int Positional_Score [2]={0,0};

	int kpos[2]; // King positions
	//int SideOfKing[2]; // King side (KINGSIDE or QUEENSIDE)

	int WhitePawnsInSqColor[2]={0,0}; // White pawns in Square color
    int BlackPawnsInSqColor[2]={0,0}; // Black pawns in Square color



	int WAttacks[64];
	int BAttacks[64];
	int row,col; //auxiliar values

	// For pawn structure. Horizontal pryections of pawn struc. with 2
	// extra columns to the rignt and left to detect out of board
	int pawn_rank[2][10]={{0,0,0,0,0,0,0,0,0,0},{7,7,7,7,7,7,7,7,7,7}};

	int prcol; // column relative a pawn_rank, not to the board;

	bool dolazyeval=true;
	int dummy,bonus;
    int i;
	int Game_Stage;



int wpawns=matsig_getP(pos->matsig[WHITE]);
int bpawns=matsig_getP(pos->matsig[BLACK]);
int wknights=matsig_getN(pos->matsig[WHITE]);
int bknights=matsig_getN(pos->matsig[BLACK]);
int wbishops=matsig_getB(pos->matsig[WHITE]);
int bbishops=matsig_getB(pos->matsig[BLACK]);
int wrooks=matsig_getR(pos->matsig[WHITE]);
int brooks=matsig_getR(pos->matsig[BLACK]);
int wqueens=matsig_getQ(pos->matsig[WHITE]);
int bqueens=matsig_getQ(pos->matsig[BLACK]);


int totalpieces=wknights+wbishops+wrooks+wqueens+
	            bknights+bbishops+brooks+bqueens;

///////////////////////////////////////////////
////// SIMPLE RECOGNIZERS. Only for safety
///////////////////////////////////////////////
if (gameDecidableDraw(pos)) return 0;
//Lazy_1_Margin[0]+=100;
///////////////////////////////////////////////
////// BITBASES PROBING
///////////////////////////////////////////////

if (egbb_is_loaded && ((totalpieces+wpawns+bpawns)<4))
	{
		int x = _NOTFOUND;
		x=probe_bitbases(pos);
		if (x!=_NOTFOUND) return x;
	}

///////////////////////////////////////////////
////// MATERIAL
///////////////////////////////////////////////

if (pos->matsig[WHITE]!=pos->matsig[BLACK]) //<-- It gets 1% speedup hehe
	{
    Material_Score[WHITE]=wpawns   * Pval- 50*(wpawns==0)+  // 90 opening 110 ending
						  wknights * Nval+                  // 400 opening 380 ending
						  wbishops * Bval+ 50 * (wbishops>1)+
						  wrooks   * Rval+
					      wqueens  * Qval;

	Material_Score[BLACK]=bpawns   * Pval- 50*(bpawns==0)+
                          bknights * Nval+
					      bbishops * Bval+ 50 * (bbishops>1)+
					      brooks   * Rval+
					      bqueens  * Qval;
	}

//dolazyeval=(totalpieces>4);


Score[WHITE]=Material_Score[WHITE];
Score[BLACK]=Material_Score[BLACK];


///////////////////////////////////////////////
////// GAME STAGE
///////////////////////////////////////////////

if ((totalpieces>=10) || ((totalpieces>=6) && ((wqueens!=0) || (bqueens!=0))))
{Game_Stage=MIDDLE;}
else
   Game_Stage=ENDING;

#ifdef LE_1
///////////////////////////////////////////////
////// FIRST LAZY EVAL CHECKING
///////////////////////////////////////////////


if (dolazyeval)
{
 if(pos->side) //Black
		dummy=Score[BLACK]-Score[WHITE];
	else
		dummy=Score[WHITE]-Score[BLACK];

if (dummy-Lazy_1_Margin[Game_Stage]>=beta)
   return beta;

if (dummy+Lazy_1_Margin[Game_Stage]<=alpha)
   return alpha;

}
#endif


/*
///////////////////////////////////////////////
////// GAME STAGE
///////////////////////////////////////////////

if ((totalpieces>=10) || ((totalpieces>=6) && ((wqueens!=0) || (bqueens!=0))))
{Game_Stage=MIDDLE;}
else
   Game_Stage=ENDING;
*/
///////////////////////////////////////////////
///// INITIATIVE
///////////////////////////////////////////////

if (Game_Stage==MIDDLE)  Score[pos->side]+= 10 ;
else Score[pos->side]+= 5;

///////////////////////////////////////////////
///// PST SCORES
///////////////////////////////////////////////

kpos[WHITE]=pos->KingPos[WHITE]^0x38;
kpos[BLACK]=pos->KingPos[BLACK]^0x38;

// Initiliazing PST_score with PST values for Kings
PST_Score[WHITE]=King_PST[Game_Stage][kpos[WHITE]];
PST_Score[BLACK]=King_PST[Game_Stage][kpos[BLACK]^0x38];

// PST Eval. Pawn Rank array build
{

 U64 WhitePawns=piecesWHITE(*pos,P);
 U64 BlackPawns=piecesBLACK(*pos,P);
 U64 WhiteKnights=piecesWHITE(*pos,N);
 U64 BlackKnights=piecesBLACK(*pos,N);
 U64 WhiteBishops=piecesWHITE(*pos,B);
 U64 BlackBishops=piecesBLACK(*pos,B);
 U64 WhiteRooks=piecesWHITE(*pos,R);
 U64 BlackRooks=piecesBLACK(*pos,R);

#ifdef PINNED_EVAL
 U64 WhitePinnedPieces=possiblePinned(pos,WHITE);
 U64 BlackPinnedPieces=possiblePinned(pos,BLACK);
#endif

 int sq;

 while (WhitePawns)
 {
  GetBitAndClear(WhitePawns,sq);
  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[WHITE]+= Pawn_PST[Game_Stage][sq];
  col=fcol(sq);
  row=frow(sq);
  if (pawn_rank[WHITE][col+1]<row) pawn_rank[WHITE][col+1]=row;

#ifdef KING_TROPISM
  if (Game_Stage==MIDDLE) Score[WHITE]+=(7-fdistance(sq,kpos[BLACK])) * KING_TROPISM;
  #endif
 }

 while (BlackPawns)
 {
  GetBitAndClear(BlackPawns,sq);
  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[BLACK]+= Pawn_PST[Game_Stage][sq^0x38];
  col=fcol(sq);
  row=frow(sq);
  if (pawn_rank[BLACK][col+1]>row) pawn_rank[BLACK][col+1]=row;

#ifdef KING_TROPISM
  if (Game_Stage==MIDDLE) Score[BLACK]+=(7-fdistance(sq,kpos[WHITE])) * KING_TROPISM;
#endif
 }

 while (WhiteKnights)
 {
  GetBitAndClear(WhiteKnights,sq);

#ifdef PINNED_EVAL
  if ((WhitePinnedPieces>>sq) & 1)
  {
  //printf("White knight pinned\n");
  PST_Score[WHITE]-=KNIGHT_PINNED_PENALTY; 
  }
#endif

  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[WHITE]+= Knight_PST[Game_Stage][sq];

#ifdef KING_TROPISM
  if (Game_Stage==MIDDLE) Score[WHITE]+=(7-fdistance(sq,kpos[BLACK])) * KING_TROPISM;
#endif
 }

 while (BlackKnights)
 {
  GetBitAndClear(BlackKnights,sq);

#ifdef PINNED_EVAL
  if ((BlackPinnedPieces>>sq) & 1)
  {
  //printf("Black knight pinned\n");
  PST_Score[BLACK]-=KNIGHT_PINNED_PENALTY; 
  }
#endif

  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[BLACK]+= Knight_PST[Game_Stage][sq^0x38];

#ifdef KING_TROPISM
  if (Game_Stage==MIDDLE) Score[BLACK]+=(7-fdistance(sq,kpos[WHITE])) * KING_TROPISM;
#endif
 }

 while (WhiteBishops)
 {
  GetBitAndClear(WhiteBishops,sq);

#ifdef PINNED_EVAL
  if ((WhitePinnedPieces>>sq) & 1)
  {
  //printf("Black knight pinned\n");
  PST_Score[WHITE]-=BISHOP_PINNED_PENALTY; 
  }
#endif

  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[WHITE]+= Bishop_PST[Game_Stage][sq];
 }
 while (BlackBishops)
 {
  GetBitAndClear(BlackBishops,sq);
  
#ifdef PINNED_EVAL
  if ((BlackPinnedPieces>>sq) & 1)
  {
  //printf("Black knight pinned\n");
  PST_Score[BLACK]-=BISHOP_PINNED_PENALTY; 
  }
#endif

  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[BLACK]+= Bishop_PST[Game_Stage][sq^0x38];
 }
 while (WhiteRooks)
 {
  GetBitAndClear(WhiteRooks,sq);

#ifdef PINNED_EVAL
  if ((WhitePinnedPieces>>sq) & 1)
  {
  //printf("Black knight pinned\n");
  PST_Score[WHITE]-=ROOK_PINNED_PENALTY; 
  }
#endif

  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[WHITE]+= Rook_PST[sq];
 }
 while (BlackRooks)
 {
  GetBitAndClear(BlackRooks,sq);

  #ifdef PINNED_EVAL
  if ((BlackPinnedPieces>>sq) & 1)
  {
  //printf("Black knight pinned\n");
  PST_Score[BLACK]-=ROOK_PINNED_PENALTY; 
  }
#endif

  sq=sq^0x38;  // to be reformated later. old Dirty format
  PST_Score[BLACK]+= Rook_PST[sq^0x38];
 }
}

/*
printf("\n");

for (i=0; i<10 ;i++)
{
 printf("%d",pawn_rank[WHITE][i]);
}

printf("\n");
for (i=0; i<10 ;i++)
{
 printf("%d",pawn_rank[BLACK][i]);
}
printf("\n");
*/


Score[WHITE] += PST_Score[WHITE];
Score[BLACK] += PST_Score[BLACK];

#ifdef LE_2
///////////////////////////////////////////////
////// SECOND LAZY EVAL CHECKING
///////////////////////////////////////////////

if (dolazyeval)
{
 if(pos->side) //Black
		dummy=Score[BLACK]-Score[WHITE];
	else
		dummy=Score[WHITE]-Score[BLACK];

if (dummy-Lazy_2_Margin[Game_Stage]>=beta)
   return beta;

if (dummy+Lazy_2_Margin[Game_Stage]<=alpha)
   return alpha;
}
#endif

///////////////////////////////////////////////
////// PAWN and PIECES EVAL
///////////////////////////////////////////////
{
 U64 WhitePawns=piecesWHITE(*pos,P);
 U64 WhitePawns2=WhitePawns;
 U64 BlackPawns=piecesBLACK(*pos,P);
 U64 BlackPawns2=BlackPawns;
 U64 WhiteRooks=piecesWHITE(*pos,R);
 U64 BlackRooks=piecesBLACK(*pos,R);
 U64 WhiteBishops=piecesWHITE(*pos,B);
 U64 BlackBishops=piecesBLACK(*pos,B);
 int WeakPawn =0 ;
 int sq;

while (WhitePawns)
{
	GetBitAndClear(WhitePawns,sq);

   	if (((BlackPawns2 | WhitePawns2)>>((sq)+8)) & 1) WhitePawnsInSqColor[Square_Color_PST[sq^0x38]]++;

    sq=sq^0x38;
	col=fcol(sq);
    row=frow(sq);
    prcol=col+1;

    WeakPawn=0;
	//Doubled
	if (pawn_rank[WHITE][prcol]>row)
	  {
		  //printf("white doubled at %d\n",sq);
		  Pawn_Score[WHITE]-=DOUBLED_PAWN_PENALTY;
		  WeakPawn++;
	  }
      // isolated
    if ((pawn_rank[WHITE][prcol+1]==0) && (pawn_rank[WHITE][prcol-1]==0))
	  {
		  // printf("white isolated at %d\n",sq);
		  Pawn_Score[WHITE]-=ISOLATED_PAWN_PENALTY;
		  WeakPawn++;
	  }
	   else
	  // backward?
	    if ((pawn_rank[WHITE][prcol+1]<row) && (pawn_rank[WHITE][prcol-1]<row))
		{
			Pawn_Score[WHITE]-=Backward_Pawn_PST[sq];
			WeakPawn++;
		}
      //passed
	 if ((pawn_rank[BLACK][prcol-1]>=row) && (pawn_rank[BLACK][prcol]>=row) && (pawn_rank[BLACK][prcol+1]>=row))
	  {
		  //printf("white passer at %d\n",thesquare);
          bonus= PASSER_BONUS+PASSER_ROW_FACTOR*(7-row)*(7-row)+fdistance(kpos[BLACK],col)*3;
		  // is passer protected ?
		  if ((pawn_rank[WHITE][prcol-1]==row+1) || (pawn_rank[WHITE][prcol+1]==row+1))
		  {
			 // printf("White passer protected at %d\n",sq);
              bonus=bonus*2;
		  }
		  Passer_Score[WHITE]+=bonus;
	  }
	  else // pawn is not passer, let's check for falange and/or semipassed
	  {

		  //Now check for semi-passer. If so bonus is the same than for passer but divided by 2
          if (((pawn_rank[BLACK][prcol-1]>=row) && (pawn_rank[BLACK][prcol]>=row)) ||
			  ((pawn_rank[BLACK][prcol+1]>=row) && (pawn_rank[BLACK][prcol]>=row)))
		  {
			 // printf("White Semipasser at : %d ",thesquare);
			  if (WeakPawn)
				  Passer_Score[WHITE]+= (PASSER_BONUS+PASSER_ROW_FACTOR*(7-row)*(7-row))/4;
			  else
     			  Passer_Score[WHITE]+= (PASSER_BONUS+PASSER_ROW_FACTOR*(7-row)*(7-row))/2;
		  }
	  }
}
while (BlackPawns)
{
	GetBitAndClear(BlackPawns,sq);

	if (((WhitePawns2 | BlackPawns2)>>((sq)-8)) & 1) BlackPawnsInSqColor[Square_Color_PST[sq^0x38]]++;


    sq=sq^0x38;
	col=fcol(sq);
    row=frow(sq);
    prcol=col+1;

	WeakPawn=0;
	if (pawn_rank[BLACK][prcol]<row)
	  {
		  Pawn_Score[BLACK]-=DOUBLED_PAWN_PENALTY;
		  WeakPawn++;
	  }
      // isolated
	  if ((pawn_rank[BLACK][prcol+1]==7) && (pawn_rank[BLACK][prcol-1]==7))
	  {
		  //printf("Black isolated at %d\n",thesquare);
		  Pawn_Score[BLACK]-=ISOLATED_PAWN_PENALTY;
		  WeakPawn++;
	  }
	   else
	  // backward?
	    if ((pawn_rank[BLACK][prcol+1]>row) && (pawn_rank[BLACK][prcol-1]>row))
		{
			//printf("Black backward at %d\n",thesquare);
			//Pawn_Score[BLACK]-=BACKWARD_PAWN_PENALTY;

			Pawn_Score[BLACK]-=Backward_Pawn_PST[sq^0x38];
			WeakPawn++;
		}
      //passed
	  if ((pawn_rank[WHITE][prcol-1]<=row) && (pawn_rank[WHITE][prcol]<=row) && (pawn_rank[WHITE][prcol+1]<=row))
	  {
		//  printf("Black passer at %d\n",thesquare);
          bonus= PASSER_BONUS+PASSER_ROW_FACTOR*row*row + fdistance(kpos[WHITE],col^0x38)*3;
		  // is passer protected ?
		  if ((pawn_rank[BLACK][prcol-1]==row-1) || (pawn_rank[BLACK][prcol+1]==row-1))
		  {
			  //printf("Black passer protected at %d\n",thesquare);
              bonus=bonus*2;
		  }
		  // add stuff
		  //printf("Black Passer bonus : %d\n",bonus);
		  Passer_Score[BLACK]+=bonus;
	  }
	  else // pawn is not passer, let's check for falange and/or semipassed
	  {
		  //Now check for semi-passer. If so bonus is the same than for passer but divided by 2
          if (((pawn_rank[WHITE][prcol-1]<=row) && (pawn_rank[WHITE][prcol]<=row)) ||
			  ((pawn_rank[WHITE][prcol+1]<=row) && (pawn_rank[WHITE][prcol]<=row)))
		  {
			//  printf("Black Semipasser at : %d %d ",thesquare,row);
			 if (WeakPawn)
			  Passer_Score[BLACK]+= (PASSER_BONUS+PASSER_ROW_FACTOR*row*row)/4;
			 else
			  Passer_Score[BLACK]+= (PASSER_BONUS+PASSER_ROW_FACTOR*row*row)/2;
		  }
	  }

}


while (WhiteRooks)
{
	GetBitAndClear(WhiteRooks,sq);
    sq=sq^0x38;
	col=fcol(sq);
    row=frow(sq);
    prcol=col+1;
	if ((pawn_rank[WHITE][prcol]>=row) || (pawn_rank[WHITE][prcol]==0))
	 {
       if ((Game_Stage==MIDDLE) && (abs(fcol(kpos[BLACK])-col)<3))
	   {
          // printf("White Rook attacking %d\n",sq);
		   Score[WHITE]+=ROOK_ATTACKING_BONUS;
	   }
       if (pawn_rank[BLACK][prcol]==7)
	   {
           //printf("White Rook in openfile at %d\n",sq);
		   Score[WHITE]+=ROOK_IN_OPEN_FILE-(7-row); // Open file
	   }

	     else
		 {
	     	//printf("White Rook in semi-openfile at %d\n",sq);
	    	Score[WHITE]+=ROOK_IN_SEMI_OPEN_FILE-(7-row); // Semi-Open file
	     }
	 }
}
while (BlackRooks)
{
	GetBitAndClear(BlackRooks,sq);
    sq=sq^0x38;
	col=fcol(sq);
    row=frow(sq);
    prcol=col+1;
	if ((pawn_rank[BLACK][prcol]<=row) || (pawn_rank[BLACK][prcol]==7))
	 {
       if ((Game_Stage==MIDDLE) && (abs(fcol(kpos[WHITE])-col)<3))
	   {
           //printf("Black Rook attacking %d\n",sq);
		   Score[BLACK]+=ROOK_ATTACKING_BONUS;
	   }
       if (pawn_rank[WHITE][prcol]==0)
	   {
		  // printf("Black Rook in openfile at %d\n",sq);
		 //  printf("White king pos %d\n",kpos[WHITE]);
		   Score[BLACK]+=ROOK_IN_OPEN_FILE-row; // Open file
	   }
	   else
	   {
		  // printf("Black Rook in semi-openfile at col %d\n",col);
		  //   printf("White king col %d\n",fcol(kpos[WHITE]));
		   Score[BLACK]+=ROOK_IN_SEMI_OPEN_FILE-row; // Semi-Open file
	   }
	}
}


while (WhiteBishops)
{
    GetBitAndClear(WhiteBishops,sq);


	if ((sq==A7) && ((BlackPawns2>>B7) & 1)) Score[WHITE]-=BISHOP_TRAPPED_A7_PENALTY; //Bishop trapped or risky
    if ((sq==H7) && ((BlackPawns2>>G7) & 1)) Score[WHITE]-=BISHOP_TRAPPED_A7_PENALTY; //Bishop trapped


	if (Game_Stage==MIDDLE)
	{
    	if ((sq==D3) && ((WhitePawns2>>D2) & 1)) Score[WHITE]-=BISHOP_E3_PENALTY; // Bishop in front of d pawn
        if ((sq==E3) && ((WhitePawns2>>E2) & 1)) Score[WHITE]-=BISHOP_E3_PENALTY;
	}


	Score[WHITE]-=WhitePawnsInSqColor[Square_Color_PST[sq^0x38]]*5; // Penalty for blocked pawns in bishop square color
	Score[WHITE]+=BlackPawnsInSqColor[Square_Color_PST[sq^0x38]]*5; // Bonus for opponent's blocked pawn in b.sq.color

}
while (BlackBishops)
{
    GetBitAndClear(BlackBishops,sq);

    if ((sq==A2) && ((WhitePawns2>>B2) & 1)) Score[BLACK]-=BISHOP_TRAPPED_A7_PENALTY; //Bishop trapped or risky
    if ((sq==H2) && ((WhitePawns2>>G2) & 1)) Score[BLACK]-=BISHOP_TRAPPED_A7_PENALTY; //Bishop trapped

	if (Game_Stage==MIDDLE)
	{
    	if ((sq==D6) && ((BlackPawns2>>D7) & 1)) Score[BLACK]-=BISHOP_E3_PENALTY; // Bishop in front of d pawn
        if ((sq==E6) && ((BlackPawns2>>E7) & 1)) Score[BLACK]-=BISHOP_E3_PENALTY;
	}

   	Score[BLACK]-=BlackPawnsInSqColor[Square_Color_PST[sq^0x38]]*5;
	Score[BLACK]+=WhitePawnsInSqColor[Square_Color_PST[sq^0x38]]*5;

}


}

Score[WHITE]+=(Pawn_Score[WHITE]*Pawn_Weight[Game_Stage])/100;
Score[BLACK]+=(Pawn_Score[BLACK]*Pawn_Weight[Game_Stage])/100;


Score[WHITE]+=(Passer_Score[WHITE]*Passer_Pawn_Weight[Game_Stage])/100;
Score[BLACK]+=(Passer_Score[BLACK]*Passer_Pawn_Weight[Game_Stage])/100;


#ifdef LE_3
///////////////////////////////////////////////
////// THIRD LAZY EVAL CHECKING
///////////////////////////////////////////////

if (dolazyeval)
{
 if(pos->side) //Black
		dummy=Score[BLACK]-Score[WHITE];
	else
		dummy=Score[WHITE]-Score[BLACK];

if (dummy-Lazy_3_Margin[Game_Stage]>=beta)
   return beta;

if (dummy+Lazy_3_Margin[Game_Stage]<=alpha)
   return alpha;

}
#endif

///////////////////////////////////////////////
////// ATTACK TABLE GENERATION and MOBILITY
///////////////////////////////////////////////

#if defined(MOBILITY_EVAL) || defined(KING_SAFETY_EVAL)
       {
       U64 map;

	 //  if ((Game_Stage==MIDDLE) || (wqueens>0) || (bqueens>0))
        {

        memset(WAttacks, 0, 64*sizeof(int));
        memset(BAttacks, 0, 64*sizeof(int));

        for (i=0; i<64 ;i++)
		{
         if(map=attacksTo(*pos,i)){  //get attacks for both colors in one go
                 WAttacks[i]+=popcnt(map & (pos)->PiecesSide[WHITE]);
                 BAttacks[i]+=popcnt(map & (pos)->PiecesSide[BLACK]);

#ifdef MOBILITY_EVAL
                 //  if (Game_Stage==MIDDLE)
                 {
                 mobility[WHITE]+=WAttacks[i]*4;
                 mobility[BLACK]+=BAttacks[i]*4;
                 }
#endif
          }
        }
	   }
#ifdef MOBILITY_EVAL

    Score[WHITE]+=mobility[WHITE];
    Score[BLACK]+=mobility[BLACK];

#endif
	   }
#endif


///////////////////////////////////////////////
////// KING SAFETY
///////////////////////////////////////////////

#ifdef KING_SAFETY_EVAL

    if ((Game_Stage==MIDDLE) || (wqueens>0) || (bqueens>0))
	{
     U64 WKingAround;
	 U64 BKingAround;
	 U64 WhitePawns=piecesWHITE(*pos,P);
     U64 BlackPawns=piecesBLACK(*pos,P);

	 int wcounter=0;
	 int bcounter=0;
	 int wshelter=0;
	 int bshelter=0;


     int IsPawn;
     if (Game_Stage==MIDDLE)
	 {
		 wcounter+=KSCounterTable[kpos[WHITE]];
		 bcounter+=KSCounterTable[kpos[BLACK]^0x38];
	 }



	 WKingAround=Kmoves(pos->KingPos[WHITE]);
     BKingAround=Kmoves(pos->KingPos[BLACK]);


	 while (WKingAround)
	 {
		 	int sq;
          	GetBitAndClear(WKingAround,sq);
	        IsPawn=((WhitePawns>>sq) & 1);
			if (IsPawn) wshelter+=2;
            if (BAttacks[sq]) wcounter+=BAttacks[sq] * 4 - WAttacks[sq];

	 }

	 while (BKingAround)
	 {
		 	int sq;
			GetBitAndClear(BKingAround,sq);
	    	IsPawn=((BlackPawns>>sq) & 1);
			if (IsPawn) bshelter+=2;
			if (WAttacks[sq]) bcounter+=WAttacks[sq] * 4 - BAttacks[sq];

	 }


     if (frow(kpos[WHITE])>4) // white king in first three rows, check the barrier attacks
	 {
	   //printf("KPos White %d",frow(kpos[WHITE]));
	   for (i=0; i<3 ;i++)
	   {
		 //  printf("%d",WKingBarrier[kpos[WHITE]^0x38][i]);

		   IsPawn=((WhitePawns >> WKingBarrier[kpos[WHITE]^0x38][i]) & 1);
		   if (IsPawn) wshelter+=1;
      if (BAttacks[WKingBarrier[kpos[WHITE]^0x38][i]])
		   wcounter+=BAttacks[WKingBarrier[kpos[WHITE]^0x38][i]]*2
				   // -(WAttacks[WKingBarrier[kpos[WHITE]^0x38][i]]>1);
		          -(WAttacks[WKingBarrier[kpos[WHITE]^0x38][i]]);


	   }
	 }
     if (frow(kpos[BLACK])<3) // Black king in first three rows, check the barrier attacks
	 {
	   for (i=0; i<3 ;i++)
	   {

		   IsPawn=((BlackPawns >> BKingBarrier[kpos[BLACK]][i]) & 1);
           if (IsPawn) bshelter+=1;
		if (WAttacks[BKingBarrier[kpos[BLACK]][i]])
		   bcounter+=WAttacks[BKingBarrier[kpos[BLACK]][i]]*2
			       // -(BAttacks[BKingBarrier[kpos[BLACK]][i]]>1);
		     		-(BAttacks[BKingBarrier[kpos[BLACK]][i]]);
	   }
	 }

	// if (wcounter<0) wcounter=0;
	// if (bcounter<0) bcounter=0;
    //if (wcounter>49) wcounter=49;
    //if (bcounter>49) bcounter=49;

	 //wcounter=max(0,min(49,wcounter));
	// bcounter=max(0,min(49,bcounter));
	 wshelter=min(9,wshelter);
	 bshelter=min(9,bshelter);

//	 wcounter=max(0,min(49,wcounter-wshelter/2));
//	 bcounter=max(0,min(49,bcounter-bshelter/2));

	 wcounter=max(0,min(49,wcounter));
	 bcounter=max(0,min(49,bcounter));

	 KingSafety[WHITE]+=-KSTable[wcounter]+Pawn_Shelter_Bonus[wshelter];
	 KingSafety[BLACK]+=-KSTable[bcounter]+Pawn_Shelter_Bonus[bshelter];

	 Score[WHITE]+=(KingSafety[WHITE]*King_Safety_Weight[Game_Stage])/100;
	 Score[BLACK]+=(KingSafety[BLACK]*King_Safety_Weight[Game_Stage])/100;

	}
#endif


/*
printf("Mat  score white : %d\n",Material_Score[WHITE]);
printf("Mat  score black : %d\n",Material_Score[BLACK]);
printf("Pst  score white : %d\n",PST_Score[WHITE]);
printf("Pst  score black : %d\n",PST_Score[BLACK]);
printf("\n");
printf("Pawn score white : %d\n",Pawn_Score[WHITE]);
printf("Pawn score black : %d\n",Pawn_Score[BLACK]);
printf("Pass score white : %d\n",Passer_Score[WHITE]);
printf("Pass score black : %d\n",Passer_Score[BLACK]);
printf("\n");
printf("Mobility white : %d\n",mobility[WHITE]);
printf("Mobility black : %d\n",mobility[BLACK]);


printf("King Safety white : %d\n",KingSafety[WHITE]);
printf("King Safety black : %d\n",KingSafety[BLACK]);

*/


#ifdef EVAL_REDUCTION
	// After adding many eval terms, sometimes positional values become too high
	// here we can adjust them.

	Positional_Score[WHITE]=Score[WHITE]-Material_Score[WHITE];
	Positional_Score[BLACK]=Score[BLACK]-Material_Score[BLACK];

	Positional_Score[WHITE]=(Positional_Score[WHITE]*(100+25*(Game_Stage==MIDDLE)))/100;
    Positional_Score[BLACK]=(Positional_Score[BLACK]*(100+25*(Game_Stage==MIDDLE)))/100;

	Score[WHITE]=Material_Score[WHITE]+Positional_Score[WHITE];
	Score[BLACK]=Material_Score[BLACK]+Positional_Score[BLACK];
#endif

if (totalpieces==0) {
Score[WHITE]+=(Score[WHITE]*50)/100;
Score[BLACK]+=(Score[BLACK]*50)/100;
}

if (totalpieces==2){
 if ((wrooks==1) && (brooks==1)) {
  Score[WHITE]-=(Score[WHITE]*25)/100;
  Score[BLACK]-=(Score[BLACK]*25)/100;
 }
 if ((wqueens==1) && (bqueens==1)) {
  Score[WHITE]-=(Score[WHITE]*25)/100;
  Score[BLACK]-=(Score[BLACK]*25)/100;
 }
}


if(pos->side) //Black
		return Score[BLACK]-Score[WHITE];
	else
		return Score[WHITE]-Score[BLACK];
}
#endif
