/**
*Buzz Chess Engine
*main.c
*
*Initializes the engine
*
*Copyright (C) 2008 Pradu Kannan. All rights reserved.
**/

#if defined(_WIN32) || defined(_WIN64)
	#define VC_EXTRALEAN
	#define WIN32_LEAN_AND_MEAN
	#define NOMINMAX
	#include <windows.h>
#else
	#include <unistd.h>
#endif

#include "eval.h"
#include "thread.h"
#include "search.h"
#include "defs.h"
#include "bitinstructions.h"
#include "movegen.h"
#include "board.h"
#include "xboard.h"
#include "consolecolors.h"
#include "hash.h"
#include "book.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "log.h"
#include "rand64.h"
#include "timemanager.h"
#include "bench.h"
#include "resultanalysis.h"
#include "nalimov.h"
#include "recog.h"
#include "egbb.h"

extern Mutex printLock; //from xboard.h
extern Mutex logLock;   //from log.h
extern bool never_ponder; //lousy

void printArrU64(U64* arr, char* name)
{
	int i;
	print("const U64 %s[64]=\n",name);
	print("{\n");
	for(i=0;i<64;i++)
	{
		if(!(i&3)) print("\t");
		print("C64(0x");
		printHex(arr[i]);
		print(")");
		if(i!=63) print(",");
		if(i && !((i+1)&3)) print("\n");
		else print(" ");
	}
	print("};\n");
}

void createELODataSet()
{
	int i;
	FILE *f = fopen("ELOdat.dat","wt");
	fprintf(f,"myELO\toppELO\tresult\n");
	for(i=0;i<10000;i++)
	{
		int R1 = rand()%1000 + 1500;
		int R2 = rand()%1000 + 1500;
		int adv = R1-R2;
		double winProb = ELOToWin(adv);
		double rnd = ((double)rand()/((double)(RAND_MAX)));
		if(rnd<=winProb)
			fprintf(f,"%d\t%d\t%d\n",R1,R2,1);
		else
			fprintf(f,"%d\t%d\t%d\n",R1,R2,0);
	}
	fclose(f);
}

int main(int argc, char *argv[])
{
	//Initialize the engine
	seed(); //seed MT RNG
	generateZobrist();
	initMoveGen();
	initSearch();
	initDistance();
	//createELODataSet();

	InitMutex(&printLock);
	InitMutex(&logLock);
	InitMutex(&rootEntryLock);
	
	srand((unsigned int)time(NULL));

	//initNalimov("G:\\3;G:\\4",8*1024*1024);
	//printf("NalimovNumPieces = %d\n",NalimovNumPieces);
	/*setboard("K/////1kr b - -");
	printBoard(gamePosition);
	{
		int UB, LB, ret;
		ret = recog(gamePosition,&LB,&UB,0);
		printf
		(
			"ret = %d\n"
			"LB  = %d\n"
			"UB  = %d\n"
			,ret,LB,UB
		);
	}*/

	//parse command-line parameters
	if(argc>=2)
	{
		if(!strcmp(argv[1],"--h") || !strcmp(argv[1],"--help")
			||!strcmp(argv[1],"-h") || !strcmp(argv[1],"-help")
			|| !strcmp(argv[1],"/h") || !strcmp(argv[1],"/help")
			|| !strcmp(argv[1],"h") || !strcmp(argv[1],"help"))
		{
			printf
			(
				/**********************************************************************************/
				"\nBook-Making Options\n"
				"-------------------\n"
				"Buzz makebook <name> <plys> <games> <razor %%> <pgnFile1> <pgnFile2> ... <pgnFileN>\n"
				"     <name>      - name/location of the book file\n"
				"     <plys>      - how many plys to use in the book\n"
				"     <games>     - add only positions with atleast <games> games\n"
				"     <razor %%>   - positions that score below razor%% will get pruned\n"
				"     <pgnFileX>  - the locations of the pgn files to use\n"
				"\n"
				"Example: Buzz makebook grand.book 30 10 40 GM2001.pgn Euwe.pgn Massive.pgn\n"
				"Example: Buzz makebook " stringer("buzz books/grand.book") " 60 20 35.5 pgns/Massive.pgn\n"
				/**********************************************************************************/
				"\nRegular Options\n"
				"---------------\n"
				"Both - and / will be accepted as option indicators.\n"
				"-bk <file>, -book <file>\n"
				"     Gives the location of the book file to use.\n"
				"     Example: -book small.book\n"
				"     Example: -book " stringer("mybooks/book with space.book") "\n"
				"     Default: default.book\n"
				"-h, -help\n"
				"     Show the command-line help.  This will only work by itself.\n"
				"-hash <megabytes>\n"
				"     Set the amout of memory to use for the hash table in megabytes.\n"
				"     Example: -hash 64\n"
				"     Default: 64 MB\n"
				"-icstalk <kibitz|whisper|adaptive|notalk>\n"
				"     Determine the manner in which analysis is given on an ics.\n"
				"     Parameter kibitz  : kibitz always\n"
				"     Parameter whisper : whisper always\n"
				"     Parameter adaptive: kibitz to computer, whisper to human\n"
				"     Parameter notalk  : no analysis information given\n"
				"     Example: -icstalk adaptive\n"
				"     Default: adaptive\n"
				"-l, -log\n"
				"     Reccomended as the first argument if used.\n"
				"     Engine will write to a log named MMDDYY_HHMMSS_log.html.\n"
				"     MMDDYY_HHMMSS represents the time of creation where\n"
				"     MM=month, DD=day, YY=year, HH=hour, MM=minute, and SS=second.\n"
				"-NalimovPath <path>, -egtb <path>\n"
				"     Give the engine the location of the Nalimov tablebases.\n"
				"     Example: -Nalimov " stringer("egtb/3men;egtb/4men;C:/Nalimov/5-men") "\n"
				"-NalimovCache <megabytes>, -cache <megabytes>\n"
				"     Cache memory for Nalimov EGTBs.\n"
				"     Example: -cache 16\n"
				"     Default: 16 MB\n"
				/**********************************************************************************/
				#ifdef ANDRES_EVAL
				"\nDirty Options\n"
				"-------------\n"
				"-degbb <Scorpio bitbase path>\n"
				"-dcache <Scorpio cache size>\n"
				#endif
			);
			return EXIT_SUCCESS;
		}
		else if(!strcmp(argv[1],"makebook"))
		{
			makebook(argc, argv);
			return EXIT_SUCCESS;
		}
		else if(*argv[1]=='-' || *argv[1]=='/') //run normally with command-line parameters
		{
			int i;
			unsigned int hashTableSize=DEFAULT_HASH_TABLE_SIZE;
			argc--;
			for(i=1; argc; argc--, i++)
			{
				char* option=argv[i];
				if(*option!='-' && *option!='/')
				{
					printError("misplaced command-line option", argv[i]);
					continue;
				}
				option++;
				if(!strcmp(option,"book") || !strcmp(option,"bk"))
				{
					argc--; i++;
					if(!argc)
					{
						printError("missing argument",argv[i-1]);
						break;
					}
					strncpy(bookPath,argv[i],sizeof(bookPath)-1);
				}
				else if(!strcmp(option,"hash"))
				{
					unsigned int mb;
					argc--; i++;
					if(!argc)
					{
						printError("missing argument",argv[i-1]);
						break;
					}
					if(sscanf(argv[i],"%u",&mb)!=1 || !mb)
					{
						argc++; i--;
						printError("bad argument",argv[i]);
						continue;
					}
					hashTableSize=mb*1024*1024;
				}
				else if(!strcmp(option,"icstalk"))
				{
					argc--; i++;
					if(!argc)
					{
						printError("missing argument",argv[i-1]);
						break;
					}
					if(!strcmp(argv[i],"kibitz"))
						icsInfoTalk = ICS_INFOTALK_KIBITZ;
					else if(!strcmp(argv[i],"whisper"))
						icsInfoTalk = ICS_INFOTALK_WHISPER;
					else if(!strcmp(argv[i],"adaptive"))
						icsInfoTalk = ICS_INFOTALK_ADAPTIVE;
					else if(!strcmp(argv[i],"notalk"))
						icsInfoTalk = ICS_INFOTALK_NOTALK;
					else
					{
						argc++; i--;
						printError("bad argument",argv[i]);
						continue;
					}
				}
				else if(!strcmp(option,"log") || !strcmp(option,"l"))
					startLogHTML();
				else if(!strcmp(option,"lousy")) never_ponder=true;
				else if(!strcmp(option,"NalimovPath") || !strcmp(option,"egtb"))
				{
					argc--; i++;
					if(!argc)
					{
						printError("missing argument",argv[i-1]);
						break;
					}
					strncpy(NalimovEGTBPath,argv[i],sizeof(NalimovEGTBPath)-1);
				}
				else if(!strcmp(option,"NalimovCache") || !strcmp(option,"cache"))
				{
					unsigned int mb;
					argc--; i++;
					if(!argc)
					{
						printError("missing argument",argv[i-1]);
						break;
					}
					if(sscanf(argv[i],"%u",&mb)!=1 || !mb)
					{
						argc++; i--;
						printError("bad argument",argv[i]);
						continue;
					}
					NalimovEGTBCache=mb*1024*1024;
				}

				//ANDRES COMMAND LINE
				#ifdef ANDRES_EVAL
				else if(!strcmp(option,"degbb"))
				{
					argc--; i++;
					if(!argc)
					{
						printError("missing argument",argv[i-1]);
						break;
					}
					strncpy(degbb,argv[i],sizeof(degbb)-1);
				}
				else if(!strcmp(option,"dcache"))
				{
					argc--; i++;
					if(!argc)
					{
						printError("missing argument",argv[i-1]);
						break;
					}
					if(sscanf(argv[i],"%d",&dcache)!=1)
					{
						argc++; i--;
						printError("bad argument",argv[i]);
						continue;
					}
				}
				#endif
				else
					printError("unknown switch",argv[i]);
			}

			//all parameters passed - perform post operations
			initHashTable(hashTableSize);
			initNalimov(NalimovEGTBPath,NalimovEGTBCache);
			if(NalimovNumPieces)
				printLogHTML(HTMLLOG_COLOR_BLUE,"Detected %d-man Nalimov EGTBs\n",NalimovNumPieces);
			else
				printLogHTML(HTMLLOG_COLOR_BLUE,"Detected no Nalimov EGTBs\n");
		}
	}
	else //no parameters passed, perform default operations
	{
		initHashTable(DEFAULT_HASH_TABLE_SIZE);
		initNalimov(NalimovEGTBPath,NalimovEGTBCache); //just for testing delete later
	}


	//Is the program opened in console?
	{
		#if defined(_WIN32) || defined(_WIN64)
			DWORD dwConsoleMode/*, dwAvailable*/;
			HANDLE hInput;
			hInput = GetStdHandle(STD_INPUT_HANDLE);
			if(!GetConsoleMode(hInput,&dwConsoleMode))
     			openedInConsole=/*PeekNamedPipe(hInput,0,0,0,&dwAvailable,0)==0*/false;
			else
				openedInConsole=true;
		#else //assume Linux/UNIX
			openedInConsole=isatty(fileno(stdin));
		#endif
	}
	
	if(openedInConsole)
		printf
		(
			PROGRAM_VERSION_STRING "\n"
			
				"Copyright (C) " COPYRIGHT_YEAR_STRING " Pradyumna Kannan, Andres Valverde & Fonzy Bluemers\n\n"

				"Dirty is provided " stringer("as-is") " without any express or implied warranty. In no event\n"
				"will the authors be held liable for any damages arising from the use\n"
				"of Dirty.\n\n"
	
		);

	#ifdef ANDRES_EVAL
	      init_bitbases();
	#endif

	xboard();
	stopLogHTML();

	DeleteMutex(rootEntryLock);
	DeleteMutex(printLock);
	DeleteMutex(logLock);
	return EXIT_SUCCESS;
}
