/**
*Buzz Chess Engine
*search.c
*
*Copyright (C) 2008 Pradu Kannan. All rights reserved.
**/
#if (defined _WIN32) || defined (_WIN64)
#include <windows.h>
#endif
#include "thread.h"
#include "timemanager.h"
#include "search.h"
#include "movegen.h"
#include "xboard.h"
#include "hash.h"
#include "eval.h"
#include "moveordering.h"
#include "consolecolors.h"
#include "log.h"
#include "recog.h"
#include <stdio.h>
#include <stdlib.h>

U64 nodes;
int iterationDepth;
int moveNumberDrawDetection;
Mutex searchStatus;

#ifdef PRINT_MAXPLY_REACHED
int maxply_reached;
#endif

#ifdef TRACE_SEARCH
bool TRACE_SEARCH_FLAG = true;
#endif

#ifdef DEBUG_PV
volatile U64 nodesd[MAX_SEARCH_DEPTH+1]; //How many nodes for each depth
volatile U64 qnodes;
volatile U64 extensions[MAX_SEARCH_DEPTH+1];
volatile U64 hashhits[MAX_SEARCH_DEPTH+1];
volatile U64 hashprobes[MAX_SEARCH_DEPTH+1];
volatile U64 cutoffs[MAX_SEARCH_DEPTH+1];
volatile U64 reductions[MAX_SEARCH_DEPTH+1];
volatile U64 OPEN_in[MAX_SEARCH_DEPTH+1];
volatile U64 CUT_in[MAX_SEARCH_DEPTH+1];
volatile U64 CUT_change[MAX_SEARCH_DEPTH+1];
volatile U64 ALL_in[MAX_SEARCH_DEPTH+1];
volatile U64 ALL_change[MAX_SEARCH_DEPTH+1];
volatile U64 Sune_first[MAX_SEARCH_DEPTH+1];
volatile U64 Sune_total[MAX_SEARCH_DEPTH+1];
volatile U64 NTS_research[MAX_SEARCH_DEPTH+1];
volatile time_ms previousDepthTime;
void resetDebugDat()
{
	int i;
	//PVnodes=0;
	for(i=0;i<=MAX_SEARCH_DEPTH;i++)
	{
		nodesd[i]=extensions[i]=hashhits[i]=hashprobes[i]
		=cutoffs[i]=reductions[i]=OPEN_in[i]=CUT_in[i]=CUT_change[i]
		=NTS_research[i]=ALL_in[i]=ALL_change[i]=Sune_first[i]
		=Sune_total[i]=U64EMPTY;
	}
}
U64 totalStat(volatile U64* arr,int plies_to_go)
{
	int i;
	U64 total=U64EMPTY;
	for(i=0;i<plies_to_go;i++)
		total+=arr[i];
	return total;
}
void printSearchStats()
{
	int i, maxply_reached;
	U64 nodes_sum=U64EMPTY;
	for(maxply_reached=1;
		nodesd[maxply_reached] && maxply_reached<=MAX_SEARCH_DEPTH+1;
		nodes_sum+=nodesd[maxply_reached], maxply_reached++);

	if(previousDepthTime)
		print("Branching Factor = %.2f\n",searchtimeUsed()/(double)previousDepthTime);
	previousDepthTime=searchtimeUsed();
	print("Qsearch Nodes/Nodes = %.2f%%\n",100*(qnodes/(double)nodes));
	print("Hash Table Usage = %f%%\n",100*HashTableUsage());
	{
		U64 sune_sum=0;
		U64 sune_first_sum=0;
		for(i=1;i<maxply_reached;i++)
		{
			sune_first_sum+=Sune_first[i];
			sune_sum+=Sune_total[i];
		}
		print("Sune%% = %.2lf%%\n",(double)100*sune_first_sum/sune_sum);
	}
	print("Depth  Node%%   Ext%% HashHit%%   Cut%% "
		"  Red%% CUTpred%% ALLpred%%  Sune%% NTSRS%%\n");
	for(i=1;i<maxply_reached;i++)
	{
		print("|%4d %6.2f %6.2f %8.2f %6.2f %6.2f %8.2f %8.2f %6.2f %6.2f\n",
			i,
			(double)100*nodesd[i]/totalStat(nodesd,maxply_reached),
			(double)100*extensions[i]/nodesd[i],
			(double)100*hashhits[i]/hashprobes[i],
			(double)100*cutoffs[i]/nodesd[i],
			(double)100*reductions[i]/nodesd[i],
			(double)100*(CUT_in[i]-CUT_change[i])/(CUT_in[i]),
			(double)100*(ALL_in[i]-ALL_change[i])/(ALL_in[i]),
			(double)100*Sune_first[i]/Sune_total[i],
			(double)100*NTS_research[i]/OPEN_in[i]
		);
	}
	print("\\-------------------------------------------------------------------------\n");


	/*print("ext=%.2f%% ",100*extensions/((double)(nodes-qnodes)));

	print("hh=%.1f%% ",100*hashhits/((double)hashprobes));
	print("cut=%.2f%% ",100*cutoffs/((double)(nodes-qnodes)));
	if(CUT_in!=U64EMPTY)
	print("CUTpred=%.2f%% ",100*(CUT_in-CUT_change_ALL)/((double)(CUT_in)));

	print("q/n=%.0f%%] ",100*qnodes/((double)nodes));

	print("Node Statistics\n");
	printU64(nodes_sum);
	print("\n");*/
}
#endif
#ifdef WINDOWS_THREADING
HANDLE Thread_Starts;
#endif
Thread searchThread;
bool searchRunning=false; //initially false
int rootMaterialBalance;
bool stoppingSearch=false;

/****************
*Node counting
****************/
U64 getNodes() { return nodes; }
void resetNodes() { nodes = U64EMPTY; }


/****************
*Search algorithm
****************/

void initSearch()
{
	InitMutex(&searchStatus);
#ifdef WINDOWS_THREADING
	Thread_Starts=CreateEvent(NULL,FALSE,FALSE,NULL);
#endif
}

//Make sure actionLock is NOT locked when calling this
void stopSearch()
{
	stoppingSearch=true;
	LockM(searchStatus);
	printLogHTML(HTMLLOG_COLOR_GREEN,"Requesting search to stop\n");
	if(searchRunning)
	{
		searchRunning=false;
		ReleaseM(searchStatus);
		Join(searchThread);
	}
	else ReleaseM(searchStatus);
	stoppingSearch=false;
}

//returns true if it made a move
THREAD_PROC_RET searchStart(void* argbwd)
{
	board pos;
	int d;
#ifdef WINDOWS_THREADING
	SetEvent(Thread_Starts); //signal that thread started ,ponderfix
#endif
	assert(argbwd!=NULL);
	{
		bwd* tempbwd=(bwd*)argbwd;
		pos=tempbwd->b;
		d=tempbwd->d;
		assert(d<=MAX_SEARCH_DEPTH);
#ifdef DEBUG_DEPTH_LIMIT
		if(d>DEBUG_DEPTH_LIMIT) d=DEBUG_DEPTH_LIMIT;
#endif
		free(argbwd);
	}

	LockM(searchStatus);
	if(searchRunning)
	{//this is actually pretty lousy ;-)
		printLogHTML(HTMLLOG_COLOR_RED,"Bug,stopping search in SearchStart() 0x%X\n",pos.hashkey);
		ReleaseM(searchStatus);
		QuitThread(false);
	}

	if(!stoppingSearch) 
		searchRunning=true;
	ReleaseM(searchStatus);

	printLogHTML(HTMLLOG_COLOR_GREEN,"Started search thread - 0x%X\n",pos.hashkey);

	if(pos.AllPieces==U64EMPTY) //don't try searching illegal positions
	{
		searchRunning=false;
		QuitThread(false);
	}

	//Result checking
	{
		moveList ml;
		genMoves(&pos,&ml);

		if(ml.moveCount==0)
		{
			searchRunning=false;
			printLogHTML(HTMLLOG_COLOR_GREEN,"Quit search thread - 0x%X\n",pos.hashkey);
			QuitThread(false);
		}
		else if(pos.fifty==100)
		{
			searchRunning=false;
			printLogHTML(HTMLLOG_COLOR_GREEN,"Quit search thread - 0x%X\n",pos.hashkey);
			QuitThread(false);
		}
		//3 rep

		else
		{
			int i;
			unsigned char count=0;
			for(i=moveNumberDrawDetection-4;i>=0;i-=2)
			{
				if(gamePosition[i].hashkey==gamePosition[moveNumber].hashkey)
				{
					count++;
					if(count>=2)
					{
						searchRunning=false;
						printLogHTML(HTMLLOG_COLOR_GREEN,"Quit search thread - 0x%X\n",pos.hashkey);
						QuitThread(false);
					}
				}
			}
		}
		//insufficient material
		if(gameDecidableDraw(&pos))
		{
			searchRunning=false;
			printLogHTML(HTMLLOG_COLOR_GREEN,"Quit search thread - 0x%X\n",pos.hashkey);
			QuitThread(false);
		}
		//if(ml.moveCount==1)
		{
			LockM(actionLock);
			if(pos.hashkey==gamePosition[moveNumber].hashkey && compSide==gamePosition[moveNumber].side)
			{
				if(ml.moveCount==1)
				{
					searchRunning=false;
					printLogHTML(HTMLLOG_COLOR_GREEN,"Making singular move from search thread %0xX\n",pos.hashkey);
					makeComputerGameMove(ml.moves->m,false);
					printLogHTML(HTMLLOG_COLOR_GREEN,"Quit search thread - 0x%X\n",pos.hashkey);
					ReleaseM(actionLock);
					QuitThread(true);
				}
			}
			ReleaseM(actionLock);

			//start searching - expect side to move to change
			{
				//iterative deepening

				hashEntry probe;
				//				bool hit;
				short score=-INF;
				int downhill=0;
				//				rootMaterialBalance=materialWHITE(&pos);
				startsearchTimer();
				newSearchID();
				clearKillers();
				clearHistory();

#ifdef CLEAR_HASH_EVERY_SEARCH
				resetHashTable();
#endif
				nodes=0;
#ifdef DEBUG_PV
				previousDepthTime=0;
				resetDebugDat();
#endif
#ifdef PRINT_MAXPLY_REACHED
				maxply_reached=0;
#endif

				for(iterationDepth=1;iterationDepth<d && searchRunning && !timeUp();iterationDepth++)
				{
					bwd* rootargs=(bwd *)malloc(sizeof(bwd));
					rootargs->b=pos;
					rootargs->d=iterationDepth;
					rootargs->alpha=-INF;
					rootargs->beta=INF;
					downhill=0;
					//ASP,simple but effective
					if(iterationDepth>=4 && abs(score)<900){
						rootargs->alpha=score-ASP;
						rootargs->beta=score+ASP;
					}

#ifdef NEW_SEARCHID_EVERY_ITERATION
					newSearchID();
#endif
#ifdef DEBUG_PV
					resetDebugDat();
#endif
#ifdef WINDOWS_THREADING
					Join(BeginThreadU(searchRoot,(void*)rootargs));
#else
					Join(BeginThread(searchRoot,(void*)rootargs));
#endif
					if(probeHash(pos.hashkey,&probe))
					{
						downhill= (score>probe.score+50);
						score=probe.score;
					}
				//} 727.5
#if 1
				if( downhill && searchRunning /*&& timeUp()*/){
					bwd* rootargs=(bwd *)malloc(sizeof(bwd));
					//allocate some more time and search a bit more
					setStartTime();
					setEndTime();
					printLogHTML(HTMLLOG_COLOR_GREEN,"extending time - bad score at root\n");			
					rootargs->b=pos;
					rootargs->d=iterationDepth;
					rootargs->alpha=-INF;
					rootargs->beta=INF;

#ifdef NEW_SEARCHID_EVERY_ITERATION
					newSearchID();
#endif
#ifdef DEBUG_PV
					resetDebugDat();
#endif
#ifdef WINDOWS_THREADING
					Join(BeginThreadU(searchRoot,(void*)rootargs));
#else
					Join(BeginThread(searchRoot,(void*)rootargs));
#endif	
					if(probeHash(pos.hashkey,&probe))
					{
						score=probe.score;
					}
				}
				} //727.5
#endif
				LockM(actionLock);
				if(gamePosition[moveNumber].hashkey==pos.hashkey && compSide==gamePosition[moveNumber].side)
				{
					int hit;
					hashEntry probe;
					hit=probeHash(pos.hashkey,&probe);
					searchRunning=false;
					printLogHTML(HTMLLOG_COLOR_GREEN,"Making a move now - 0x%X\n",pos.hashkey);
					if(hit)
						makeComputerGameMove(extractMove(probe.data),false);
					else
					{
						print("Error (no move in hash table): "__FILE__ " %d\n",__LINE__);
						makeComputerGameMove(ml.moves->m,false);
					}
					printLogHTML(HTMLLOG_COLOR_GREEN,"Quit search thread - 0x%X\n",pos.hashkey);
					ReleaseM(actionLock);
					QuitThread(true);
				}
				ReleaseM(actionLock);
				searchRunning=false;
				printLogHTML(HTMLLOG_COLOR_GREEN,"Quit search thread - 0x%X\n",pos.hashkey);
				QuitThread(false);
			}
		}
	}
	//GCC will complain about the control reaching here but no worries here
}

/*************************************
*Search Algorithm: Alphabeta on a move
*************************************/

THREAD_PROC_RET searchRoot(void* bwdargs)
{
	moveList ml;
	int score=-INF;
	move bestMove=NULLMOVE;
	unsigned int i;
	hashEntry probe;
	move hashMove=NULLMOVE;
	U64 drawTable[MAX_SEARCH_DEPTH+1];
#ifdef PRINT_MOVENODES
	U64 mynodes=nodes;
#endif

	board pos=((bwd*)bwdargs)->b;
	int depth=((bwd*)bwdargs)->d;
	int alpha=((bwd*)bwdargs)->alpha;
	int beta=((bwd*)bwdargs)->beta;

	int nodeType = OPEN_NODE;

	free(bwdargs);

	assert(alpha<beta);

	genMoves(&pos,&ml);
	if(probeHash(pos.hashkey,&probe))
		hashMove=extractMove(probe.data);
	orderMoves(&pos,&ml,0,hashMove);
	sortMoveList(&ml);

	*drawTable=pos.hashkey;

	assert(ml.moveCount>0);
	for(i=0;i<ml.moveCount;i++)
	{
		int tempVal;
		int lowerbound;

		lowerbound=max(alpha,score);
		if(lowerbound+1==beta || nodeType == OPEN_NODE){
			 tempVal=-search(pos,ml.moves[i].m,-beta,-lowerbound,depth-1,1,drawTable,nodeType);
			 //check for APS and if we got out of a-b range
			 if(i==0 && alpha>-INF && (tempVal <= alpha || tempVal>beta)){
				//out of bounds
				//continue with -INF,INF window
				printLogHTML(HTMLLOG_COLOR_GREEN,"Root window too small,researching with full window\n");
				 beta=INF;
				 alpha=-INF;
				tempVal=-search(pos,ml.moves[i].m,-beta,-alpha,depth-1,1,drawTable,nodeType);
			 }
		}
		else
		{
			tempVal=-search(pos,ml.moves[i].m,-(lowerbound+1),-lowerbound,depth-1,1,drawTable,nodeType);
			if(tempVal > lowerbound && tempVal < beta)
			{
				tempVal=-search(pos,ml.moves[i].m,-beta,-(tempVal-1),depth-1,1,drawTable,OPEN_NODE);
			}
		}

		nodeType = ALL_NODE;
#ifdef PRINT_MOVENODES
		{
			sanString s;
			moveToSAN(&pos,ml.moves[i].m,s);
			print("%s ",s);
			printU64(nodes-mynodes);
			print("\n");
			mynodes=nodes;
		}
#endif
#ifdef PRINT_MOVESCORES
		{
			sanString s;
			moveToSAN(&pos,ml.moves[i].m,s);
			if(tempVal>beta)
				print("%s++ (%d,%d) %d\n",s,score,beta,tempVal);
			else
				print("%s (%d,%d) %d\n",s,score,beta,tempVal);
		}
#endif
		if(tempVal>score)
		{
			bestMove=ml.moves[i].m;
			score=tempVal;
			storeRootHash(pos.hashkey,bestMove,-INF,-INF+1,score,depth,0); //lowerbound
#ifndef DEBUG_PV
			if(post) //printing PV (depth score time nodes PV)
			{
				print("%d ",depth);
				if(score<0) CC_Red();
				if(score==0) CC_Blue();
				if(score>0) CC_Green();
				print("%d ",score);
				CC_Normal();
				print("%u ",searchtimeUsed()/10);
				printU64(nodes);
				print(" ");
				if(ponder && compSide==gamePosition[moveNumber].xside)
				{
					sanString ss;
					moveToSAN(gamePosition+moveNumber,ponderMove,ss);
					if(inCheck(pos,pos.side)) print("(%s+) ",ss);
					else print("(%s) ",ss);
				}

				printPV(pos);
				print("\n");
			}
#endif
			if(score>=beta){
				storeRootHash(pos.hashkey,bestMove,alpha,beta,score,depth,0);
				break;
			}
		}
	}
	storeRootHash(pos.hashkey,bestMove,alpha,beta,score,depth,0);
#ifdef DUMP_REAL_PV
	hashPV(&pv,pos,alpha,beta,score);
#endif
	if(post) //printing PV (depth score time nodes PV)
	{
		CC_INormal();
		print("%d ",depth);
		if(score<0) CC_IRed();
		if(score==0) CC_IBlue();
		if(score>0) CC_IGreen();
		print("%d ",score);
		CC_INormal();
		print("%u ",searchtimeUsed()/10);
		printU64(nodes);
		print(" ");
		if(ponder && compSide==gamePosition[moveNumber].xside)
		{
			sanString ss;
			moveToSAN(gamePosition+moveNumber,ponderMove,ss);
			if(inCheck(pos,pos.side)) print("(%s+) ",ss);
			else print("(%s) ",ss);
		}
		printPV(pos);
		print("\n");
#ifdef DEBUG_PV
		printSearchStats();
#endif
		CC_Normal();
	}
	QuitThread(score);
}

int search(board pos,move m, int alpha, int beta, int depth, int dfr, U64* drawTable, int p_nodeType)
{
	hashEntry probe;
	bool hashHit = false;
	moveList ml;
	int bestVal=-INF;
	move bestMove=NULLMOVE;
	bool extended=false;
	U64 checkers=U64EMPTY;
	int mvsdone=0;
	int see_m=0;
	int nodeType = -(p_nodeType);//;childNodeType(p_nodeType);
#ifdef DEBUG_PV
	int enterType = nodeType;
#endif

#ifdef PRINT_MAXPLY_REACHED
	if(dfr>maxply_reached)
		print("Max ply reached = %d\n",maxply_reached=dfr);
#endif

	assert(alpha<beta);

	nodes++;

#ifdef DEBUG_PV
	nodesd[dfr]++;
#endif
	if(!(nodes&1023) && (timeUp() || searchRunning==false)) QuitThread(false);
	//moved way up,because pos will be pointing to nonsense if we overstep
	if(dfr>=MAX_SEARCH_DEPTH) return eval(&pos,alpha,beta);
#ifdef TRACE_SEARCH
	if(TRACE_SEARCH_FLAG)
	{
		int i;
		sanString s;
		for(i=0;i<dfr-1;i++) print(" ");
		moveToSAN(&pos,m,s);
		print("%s (%d, %d) ",s,alpha,beta);
		printU64(nodes);
		print("\n");
		MilliSleep(TRACE_SEARCH);
	}
#endif
	see_m=SEE(&pos,m);
	makemove(&pos,m);

	//test stop conditions here
	if(pos.fifty==100)//fifty move rule
	{
		U64 checkers;
		if(!(checkers=inCheck(pos,pos.side)))
			return 0;
		else //make sure we arn't mated
		{
			moveList ml;
			//this case is rare, so hardly any loss in speed to generate moves here
			genEvasions(&pos,&ml,checkers);
			if(ml.moveCount==0)
				return -MATE+dfr;
			else
				return 0;
		}
	}
	//do repetition here
#ifdef NULLMOVE_NOREPCHECK
	if(stripMove(m)!=NULLMOVE)
#endif
	{
		int j;
		for(j=dfr-4;j>=0;j-=2)
			if(pos.hashkey==drawTable[j])
				return 0;
		j+=moveNumberDrawDetection;
		while(j>=0)
		{
			if(pos.hashkey==gamePosition[j].hashkey)
				return 0;
			j-=2;
		}
	}
	drawTable[dfr]=pos.hashkey;

#ifdef MATE_DISTANCE_PRUNING
	{
		//The maximum mate score that can be attained by 
		//the current player
		//MATE-(dfr+1)
		int mate_score_bound=MATE-(dfr+1);
		if(alpha>=mate_score_bound) return mate_score_bound;
		if(beta>mate_score_bound) beta=mate_score_bound;

		//The maximum mate score that the opponent can
		//have (mate right now)
		mate_score_bound=-(MATE-dfr);
		if(mate_score_bound>=beta) return mate_score_bound;
		if(mate_score_bound>alpha) alpha=mate_score_bound;
	}
#endif
	if(gameDecidableDraw(&pos)) return 0;
	checkers=inCheck(pos,pos.side);

	if(checkers) //check extension
	{
		depth++; 
		assert(extended==false);
		extended=true;
	}

#ifdef PAWN_TO_7TH_EXTENSION
	/* extractExtra(m)==MO_PASSER -> extractPiece==P,works correct with captures to 7th/2nd rank */
	//if(extractPiece(m)==P && toBit(extractTo(m))&C64(0x00FF00000000FF00))
	if((extractExtra(m)&MO_PASSER) && toBit(extractTo(m))&C64(0x00FF00000000FF00))
	{
		if(!extended){
			depth++;
			assert(extended==false);
			extended=true;
		}
	}
#endif
	if(depth<=0)//
		return qsearchRoot(&pos,alpha,beta,dfr,checkers);//LF

#ifdef KING_PAWN_ENDING_EXTENSION	
	if(!extended && extractCapture(m)!=E && extractCapture(m)!=P
		&& pos.Pieces[R]==U64EMPTY && pos.Pieces[Q]==U64EMPTY && pos.Pieces[B]==U64EMPTY && pos.Pieces[N]==U64EMPTY)
	{
		depth+=KING_PAWN_ENDING_EXTENSION;
		assert(extended==false);
		extended=true;
	}
#endif

#ifdef PASSER_EXTENSION
	if(!extended && extractExtra(m)==MO_PASSER && nodeType==OPEN_NODE)
	{
		depth++;
		assert(extended==false);
		extended=true;
	}
#endif
	hashHit=probeHash(pos.hashkey,&probe);

#ifdef DEBUG_PV
	hashprobes[dfr]++;
#endif
	if(hashHit)
	{
		int tt_depth=probe.depth;
		int tt_flag=extractFlag(probe.data);
		int tt_score=probe.score;
		see_m=probe.score;
		if(see_m<=-MATE_MIN) see_m=0;//TUT

#ifdef DEBUG_PV
		hashhits[dfr]++;
#endif

#ifdef USE_HASH
		//Regular Hash Usage by Depth	
		if(tt_depth>=depth 
			|| (abs(tt_score)>=MATE_MIN)) /* nice trick, makes MATE_DISTANCE_PRUNING almost useless(?)*/
		{
			if(tt_score>=MATE_MIN) tt_score-=dfr;
			else if(tt_score<=-MATE_MIN) tt_score+=dfr;

			switch(tt_flag)
			{
			case HASHFLAG_EXACT:
				if(tt_depth>=depth)//mate scores from hash have to be compared with alpha,beta
					return tt_score;
			case HASHFLAG_LOWER_BOUND:
				{
					if(tt_score>=beta) 
					{
#ifdef DEBUG_PV
						cutoffs[dfr]++;
#endif
						return tt_score;
					}
#ifdef UPDATE_BOUNDS
					if(tt_score>alpha && tt_depth>=depth )
						alpha=tt_score;
#endif
					break;
				}
			case HASHFLAG_UPPER_BOUND:
				{
					if(tt_score<=alpha)
					{
#ifdef DEBUG_PV
						cutoffs[dfr]++;
#endif
						return tt_score;
					}
#ifdef UPDATE_BOUNDS
					if(tt_score<beta && tt_depth>=depth )
						beta=tt_score;
#endif
					break;
				}
			}
		}
#endif //USE_HASH
	}

#ifdef NALIMOV
	if(depth>=NALIMOV_DEPTH )
		//check recognizer
		//probe just after a root move 
		//or if depth is big enough
		//but in the second case,check if move is cap or promotion
	{
		int LB, UB; //Lowerbound upperbound)

		if(recog(&pos,&LB,&UB,dfr))
		{
			//update bounds first
			//better not
			//seems visual C screws the pointers up
			//even if recog returned false
			//...................
			//seems to be ok in debug mode though..

			//if(LB>alpha) alpha=LB;
			//if(UB<beta) beta=UB;

			//then check for cutoffs
			if(LB>=beta) return LB;
			if(UB<=alpha) return UB;
		}
	}
#endif

	assert(alpha<beta);
#ifdef NULLMOVE_PRUNING
	if(
		(alpha+1==beta) && 
#ifdef NULLMOVE_NO_ALLNODE
		nodeType!=ALL_NODE &&
#endif
#ifdef NULLMOVE_NO_OPENNODE
		nodeType!=OPEN_NODE &&
#endif
#ifdef NULLMOVE_NO_MULTI_NULLMOVE
		!extractSFNull(pos.SF) &&
		//(stripMove(m)!=NULLMOVE &&
#endif
#ifdef NULLMOVE_NO_BETAMATE
		beta<MATE_MIN &&
#endif
#ifdef NULLMOVE_DEPTH_LEFT_LIMIT
		depth>=NULLMOVE_DEPTH_LEFT_LIMIT &&
#endif
#ifdef TOTALPIECES_NULLMOVE_OFF
		matsig_getAll(pos.matsig) > TOTALPIECES_NULLMOVE_OFF &&
#endif
		!checkers && 
		//popcnt(pos.PiecesSide[pos.side]&~(pos.Pieces[P]|pos.Pieces[K]))>PIECES_NULLMOVE_OFF
		matsig_getnull(pos.matsig[pos.side])>PIECES_NULLMOVE_OFF
#ifdef NULLMOVE_NO_QS_ABOVE_BETA
		&& eval(&pos,alpha,beta)>=beta
#endif
		)
	{ 
		int tempVal=-search(pos,NULLMOVE,-beta,-(beta-1),depth-1-adaptiveR(&pos,depth),dfr+1,drawTable,nodeType);
		if(tempVal>=beta)
		{
#ifdef DEBUG_PV
			cutoffs[dfr]++;
#endif
			return beta;
		}
#ifdef MATE_THREAT_EXTENSION
		if(
			tempVal<alpha && //!
			tempVal<=-MATE+5 //tempVal<-MATE_MIN
			&& !extended 
			&& depth<MATE_THREAT_EXTENSION_DEPTH_LIMIT
			&& dfr<MATE_THREAT_EXTENSION_DFR_LIMIT
			&& SEE(&pos,m)<0 //!
			)
		{
#ifdef DEBUG_PV
			extensions[dfr]++;
#endif
			depth++;
			extended=true;
		}
#endif
	}
#endif
	//skipnull:
	genMoves(&pos,&ml);
	if(ml.moveCount==0)
	{
		if(checkers)
			return -MATE+dfr; //mate
		else return 0; //stalemate
	}

#ifdef SINGULAR_REPLY_EXTENSIONS
	if(ml.moveCount==1)
	{
#ifdef DEBUG_PV
		extensions[dfr]++;
#endif
#ifndef SINGULAR_REPLY_EXTEND_ALWAYS
		if(!extended)
#endif
		{
			extended=true;
			depth++;
		}
	}
	else
#endif
		orderMoves(&pos,&ml,dfr,extractMove(probe.data));
#ifdef DEBUG_PV
	if(extended) extensions[dfr]++;
#endif

#ifdef DEBUG_PV
	switch(nodeType)
	{
	case CUT_NODE: CUT_in[dfr]++; break;
	case ALL_NODE: ALL_in[dfr]++; break;
	case OPEN_NODE: OPEN_in[dfr]++; break;
	}
#endif
	//Move loop
	{

#ifndef  NDEBUG
		int mvs=ml.moveCount;
#endif
		move currMove;
		int moveorderscore;//...because nextmove() destroys ml.moves[n].score...
		//currMove-currVal
		//BestMove-BestScore
		while( (currMove=nextmove(&ml,&moveorderscore)) !=NULLMOVE)
		{
			int currVal;//value for current move,was tempVal
			int lowerbound=max(alpha,bestVal);
			int extendDepth=0;
			int reduction=0;
			
			mvsdone++;//move counter for stalemates
			reduction=0;//moved into loop

#ifdef RECAPTURE_EXTENSIONS
			extendDepth=0;
			if(!extended 
#ifdef RECAPTURE_OPEN_NODE_ONLY
				&& (p_nodeType==OPEN_NODE /*|| nodeType==CUT_NODE*/)
#endif
				&& alpha+1!=beta
				&& extractCapture(m)!=E
				&& extractTo(m)==extractTo(currMove)
				&& moveorderscore ==MOVEORDERING_CAPTURE
				) //recapture extension
			{extendDepth=1;}
#endif

			//LMP
			//prune late (noncapturing/promoting)moves
			if(mvsdone>LMR_MOVENUMBER &&  depth<=(LMR_DEPTHLEFT) && 
				//!checkers &&
				extractCapture(currMove)==E
				&& extractPromotion(currMove)==E
				&& !extended && !extendDepth 
				&& (!(extractExtra(currMove)&MO_CHECK))////
				&& alpha+1==beta
				&& (was_searched(currMove,pos.side))
				&& bestVal<MATE_MIN
				&& bestVal>-MATE_MIN
				&& pos.fifty<97
				&& (!(extractExtra(currMove)&MO_PASSER))
				){
					assert(bestVal!=INF);
					continue; 
			}

#ifdef LMR
			if( mvsdone>LMR_MOVENUMBER
#ifdef LMR_NOT_IN_CHECK
				&& (checkers==0)
#endif
#ifdef LMR_NOT_CHECKING
				&&(!(extractExtra(currMove)&MO_CHECK))
#endif
#ifdef LMR_DEPTHLEFT
				&& ( depth>LMR_DEPTHLEFT)
#endif
#ifdef LMR_NOEXTEND
				&& !extended && !extendDepth
#endif
#ifdef LMR_NO_OPENNODE
				&& nodeType != OPEN_NODE
#endif
#ifdef LMR_NORMALORDER
				&& moveorderscore <MOVEORDERING_CAPTURE/*MOVEORDERING_GOODMOVE*/
#endif
#ifdef LMR_NOTCAPTURES
				&& extractCapture(currMove)==E
				&& extractPromotion(currMove)==E
#endif
#ifdef LMR_NOTPASSERS
				&& ((extractExtra(currMove)!=MO_PASSER) || 
				(matsig_getnull(pos.matsig[pos.side])>PIECES_NULLMOVE_OFF))
#endif
				&& was_searched(currMove,pos.side)
				)
			{
				reduction=LMR_R;
				//double reductions:
				if(
					(alpha+1==beta) && 
					!checkers &&
					!extended && !extendDepth &&
					extractCapture(currMove)==E &&
					extractPromotion(currMove)==E &&
					!(extractExtra(currMove)&MO_PASSER) &&
					!(extractExtra(currMove)&MO_CHECK) 
					){
						reduction++;
				}
			}

#endif
			updatecount(currMove,pos.side);
#if 0
			if((lowerbound+1==beta) || (nodeType == OPEN_NODE )){
				currVal=-search(pos,currMove,-beta,-lowerbound,depth-1+extendDepth,dfr+1,drawTable,nodeType);
			}
			else if(nodeType == ALL_NODE)
			{   
				currVal=-search(pos,currMove,-(lowerbound+1),-lowerbound,depth-1+extendDepth-reduction,dfr+1,drawTable,nodeType);
				if(currVal > lowerbound && currVal < beta)
				{
#ifdef DEBUG_PV
					if(alpha+1!=beta) NTS_research[dfr]++;
#endif
					currVal=-search(pos,currMove,-beta,-(currVal-1),depth-1+extendDepth-reduction,dfr+1,drawTable,OPEN_NODE);
				}
			}
			else
			{
				currVal=-search(pos,currMove,-beta,-(beta-1),depth-1+extendDepth-reduction,dfr+1,drawTable,nodeType);
				if(currVal > lowerbound && currVal < beta)
				{
					currVal=-search(pos,currMove,-(currVal+1),-lowerbound,depth-1+extendDepth-reduction,dfr+1,drawTable,OPEN_NODE);
				}
			}
#else
			//simplified

			if(bestVal==-INF)
				currVal=-search(pos,currMove,-beta,-lowerbound,depth-1+extendDepth,dfr+1,drawTable,nodeType);
			else
			{   
				currVal=-search(pos,currMove,-(lowerbound+1),-lowerbound,depth-1+extendDepth-reduction,dfr+1,drawTable,nodeType);
				if(currVal > lowerbound  && currVal < beta)
				{
#ifdef DEBUG_PV
					if(alpha+1!=beta) NTS_research[dfr]++;
#endif
					currVal=-search(pos,currMove,-beta,-(currVal-1)/*-lowerbound*/,depth-1+extendDepth-reduction,dfr+1,drawTable,OPEN_NODE);
				}
				if(currVal > lowerbound && reduction!=0)
					currVal=-search(pos,currMove,-beta,-lowerbound,depth-1+extendDepth,dfr+1,drawTable,nodeType);
			}

#endif
			if(nodeType==OPEN_NODE) nodeType=CUT_NODE;
			else
				nodeType = ALL_NODE;
#ifdef TRACE_SEARCH
			if(TRACE_SEARCH_FLAG)
			{
				int index;
				for(index=0;index<dfr;index++) print(" ");
				print("Value = %d\n",currVal);
				MilliSleep(TRACE_SEARCH);
			}
#endif

			if(currVal>bestVal)
			{
				if(currVal>=beta)
				{

#ifdef USE_NULL_KILLER
					if(stripMove(m)==NULLMOVE)
						storeNullKiller(stripMove(currMove),dfr);
					//else
#endif
#ifdef TRACE_SCORES
					if(TRACE_SEARCH_FLAG)
						print("beta cut (%d,%d) %d\n", max(alpha,bestVal), beta, currVal);
#endif
					storeGoodMove(stripMove(currMove),depth,pos.side,dfr);
					storeHash(pos.hashkey,currMove,alpha,beta,currVal,depth,dfr);
#ifdef DEBUG_PV
					cutoffs[dfr]++;
					if(enterType==ALL_NODE) ALL_change[dfr]++;
					if(!i) Sune_first[dfr]++;
					Sune_total[dfr]++;
#endif
#ifdef RESHUFFLE
					//reshuffle 
					{
						unsigned int n;
						if(extractCapture(currMove)==E && (extractPromotion(currMove)==E))
							for(n=0;n<ml.moveCount;n++){
								if(ml.moves[n].m!=currMove && ml.moves[n].score==-(MOVEORDERING_HASHMOVE)){
									if(extractCapture(ml.moves[n].m)==E && (extractPromotion(ml.moves[n].m)==E))
										AdjustMoveScore(ml.moves[n].m,depth,pos.side);
								}
							}
					}
#endif
					return currVal;
				}
				bestMove=currMove;
				bestVal=currVal;
#ifdef TRACE_SCORES
				if(TRACE_SEARCH_FLAG)
					print("val>score (%d,%d) %d\n", max(alpha,bestVal), beta, currVal);
#endif
			}
		}
		//check if nextmove() didn't drop something
		assert(mvs==mvsdone);
	}
	//if(mvsdone==0) return 0; //stalemate
	assert(mvsdone!=0);
	storeHash(pos.hashkey,bestMove,alpha,beta,bestVal,depth,dfr);
	/*if(bestVal>=alpha) //History heuristic works better only in beta cutoffs
	storeHistory(bestMove,depth,pos.side);*/
	return bestVal;
}

/*****************
*Quiescence Search
*****************/

int qsearchRoot(board* pos, int alpha, int beta, int dfr,U64 checkers)
{
	int score=-INF;

#ifdef DO_SYMMETRY_CHECK
	symmetryCheck(pos);
#endif

	if(dfr>=MAX_SEARCH_DEPTH) return eval(pos,alpha,beta);
	{
		moveList ml;
		unsigned int i;

		if(!checkers){
			score=eval(pos,alpha,beta);
			if(score>=beta)
				return score;
#ifdef QSEARCH_CHECKS
			genQCMoves(pos,&ml);
#else
			genQMoves(pos,&ml);
#endif
			orderQMoves(pos,&ml);
		}
		else{
			genEvasions(pos,&ml,checkers);
			if(ml.moveCount==0)return -MATE+dfr;
		}
		
		for(i=0;i<ml.moveCount;i++)
		{
			int tempVal;

			if(!checkers && ml.moves[i].score<0) break;

			tempVal=-qsearch(*pos,ml.moves[i].m,-beta,-max(alpha,score),dfr+1);
			if(tempVal>score)
			{
				if(tempVal>=beta) return tempVal;
				score=tempVal;
			}
		}
	}
	return score;
}

int qsearch(board pos, move m, int alpha, int beta, int dfr)
{
	int score;

#ifdef COUNT_QSEARCH_NODES
	nodes++;
#endif
#ifdef DEBUG_PV
	qnodes++;
#endif
	if(!(nodes&1023) && (timeUp() || searchRunning==false)) QuitThread(false);
	if(dfr>=MAX_SEARCH_DEPTH) return eval(&pos,alpha,beta);		
	makemove(&pos,m);
#ifdef TRACE_SEARCH
	if(TRACE_SEARCH_FLAG)
	{
		int i;
		sanString s;
		print("q");
		for(i=1;i<dfr-1;i++) print(" ");
		moveToSAN(&pos,m,s);
		print("%s (%d, %d) ",s,alpha,beta);
		printU64(nodes);
		print("\n");
		MilliSleep(TRACE_SEARCH);
	}
#endif

#ifdef MATE_DISTANCE_PRUNING
	{
		//The maximum mate score that can be attained by 
		//the current player
		//MATE-(dfr+1)
		int mate_score_bound=MATE-(dfr+1);
		if(alpha>=mate_score_bound) return mate_score_bound;
		if(beta>mate_score_bound) beta=mate_score_bound;

		//The maximum mate score that the opponent can
		//have (mate right now)
		mate_score_bound=-(MATE-dfr);
		if(mate_score_bound>=beta) return mate_score_bound;
		if(mate_score_bound>alpha) alpha=mate_score_bound;
	}
#endif

#ifdef DO_SYMMETRY_CHECK
	if(!symmetryCheck(&pos))
	{
		puts("Symmetry Check Failed!");
		printBoard(&pos);
		fen(&pos,false);
		system("Pause");
	}
#endif

	{
		moveList ml;

		//check evasions
		U64 checkers=inCheck(pos,pos.side);
		if(checkers)
		{
			unsigned int i;
			int tempVal;
			genEvasions(&pos,&ml,checkers);
			if(ml.moveCount==0)
				return -MATE+dfr;
			score=-INF;
			for(i=0;i<ml.moveCount;i++)
			{
				tempVal=-qsearch(pos,ml.moves[i].m,-beta,-max(alpha,score),dfr+1);
				if(tempVal>score)
				{
					if(tempVal>=beta) return tempVal;
					score=tempVal;
				}
			}
			return score;
		}
		score=eval(&pos,alpha,beta); //stand-pat
		if(score>=beta)
			return score;

		genQMoves(&pos,&ml);
		orderQMoves(&pos,&ml);

		//iterate through moves
		{
			unsigned int i;
			int tempVal;

			for(i=0;i<ml.moveCount;i++)
			{
				if(ml.moves[i].score<0) break;

				tempVal=-qsearch(pos,ml.moves[i].m,-beta,-max(alpha,score),dfr+1);
				if(tempVal>score)
				{
					if(tempVal>=beta) return tempVal;
					score=tempVal;
				}
			}
			return score;
		}
	}
}


int adaptiveR(const board* pos, const int depth)
{
#ifdef NULLMOVE_ENSURE_OPPMOVE
	if(depth<=3) return 1;
	if(depth==4) return 2;
#endif
#ifdef NULLMOVE_ADAPTIVE_R
	{
		U64 pieces=pos->PiecesSide[pos->side]&~(pos->Pieces[P]|pos->Pieces[K]);
		int piecesCurrentSide = popcnt(pieces);
		return 2 + ((depth) > (6 + ((piecesCurrentSide<3)?2:0)));
	}
#endif
#ifdef NULLMOVE_CONST_R
	return NULLMOVE_CONST_R;
#endif
}

int childNodeType(const int previous_type)
{
	//return previous_type;
	switch(previous_type)
	{
	case ALL_NODE: return CUT_NODE;
	case CUT_NODE: return ALL_NODE;
	default: return OPEN_NODE;
	}
}

/****************
*Perft and Divide
*****************/

U64 perft(board pos, move m, int d)
{
	U64 currnodes;
	unsigned int i;
	moveList ml;

	makemove(&pos,m);
#ifdef PERFT_HASH_TEST
	{
		perftEntry* probe=probePerftHash(pos.hashkey);
		if(probe!=NULL && extractPerftDepth(probe->data)==d)
			return extractPerftNodes(probe->data);
	}
#endif

	genMoves(&pos,&ml);
	if(d-1==0)
		return ml.moveCount;
	currnodes=0;
	for(i=0;i<ml.moveCount;i++)
		currnodes+=perft(pos,ml.moves[i].m,d-1);
#ifdef PERFT_HASH_TEST
	storePerftHash(pos.hashkey,currnodes,d);
#endif
	return currnodes;
}

//takes in a bwd and free's data
DTHREAD_PROC_RET perftStart(void* args)
{
	bwd* temp;
	board pos;
	moveList ml;
	unsigned int d;
	unsigned int i;
	int t0;
	U64 currnodes=U64EMPTY;

	assert(args!=NULL);
	resetHashTable();

	temp=(bwd*)(args);
	pos=temp->b;
	d=temp->d;
	free(temp);

	nodes=0;
	t0=getms();
	genMoves(&pos,&ml);
	for(i=0;i<ml.moveCount;i++)
		currnodes+=perft(pos,ml.moves[i].m,d-1);
#ifdef PERFT_HASH_TEST
	print("\nPerft %d (with lousy hashing scheme)\n",d);
#else
	print("\nPerft %d\n",d);
#endif
	print("Time: %d ms\n",(getms()-t0));
	print("Leaf Nodes: ");
	printU64(currnodes);
	print("\n");
}

void divide(board pos, int d)
{
	unsigned int i;
	moveList ml;

	genMoves(&pos,&ml);
	for(i=0;i<ml.moveCount;i++)
	{
		coordString mstr;
		moveToCoords(&pos, ml.moves[i].m,mstr);
		print("%s ",mstr);
		printU64(perft(pos,ml.moves[i].m,d-1));
		print("\n");
	}
}
